'''
pipeline.py

runs all models (or a subset) for a specified dataset.
parameters are customizable. This is a work in progress
and should be modularized such that pieces can be put in
and taken out with ease.

available models:
    - lr
    - hmm by sgd
    - gru
    - gru hmm

available cost functions
    - obs-out trade-off

available regularization
    - redun
    - kdropout
    - l1

'''
from __future__ import print_function

import os
import sys
import pdb
import shutil
import cPickle
import numpy as np
from datetime import datetime
from rlstm import models_util as util
from rlstm.scores import get_log_sparsity, get_accuracy, get_auc
from rlstm.common_util import (flatten_to_2d, merge_two_dicts, pp_json, map_3d_to_2d)
from copy import copy


def get_relevant_args(func, unused_set=None):
    unused_set = set(unused_set)
    args = set(func.func_code.co_varnames[:func.func_code.co_argcount])
    return list(args - unused_set)


class RLSTMShareWeights(object):
    ''' Define a system to transfer weights from a previous node
        to a latter node at pipeline execution.
    '''

    def __init__(self,
                 parent_name,
                 parent_weights_attr='init_weights',
                 parent_weights_idx=None):
        self.parent_name = parent_name
        self.parent_weights_attr = parent_weights_attr
        self.parent_weights_idx = parent_weights_idx

    def get_parent_weights(self, pipeline):
        weights = pipeline.get_node(self.parent_name).params['trained_weights']
        if not self.parent_weights_idx is None:
            weights = weights[self.parent_weights_idx.astype(bool)]
        return weights


class RLSTMNode(object):

    ''' Doubly linked set for declaring execution pieces.
    return list(args - irrelevant)
    '''
    def __init__(self,
                 model_name,
                 model_class,
                 index,
                 prev,
                 next,
                 model_params=None):

        self.name = model_name
        self.index = index
        self.prev = prev
        self.next = next
        self.type = model_class

        if model_class == 'lr':
            func = util.train_lr
        elif model_class == 'hmm-sgd':
            func = util.train_hmm_sgd
        elif model_class == 'gru':
            func = util.train_gru
        elif model_class == 'mlp':
            func = util.train_mlp
        elif model_class == 'gru-hmm':
            func = util.train_gru_hmm
        elif model_class == 'gru-hmm-tree':
            func = util.train_gru_hmm_tree
        self.func = func

        params = self.default_params()
        if model_params:
            for key in model_params:
                params[key] = model_params[key]

        self.params = params
        self.unused_params = ['obs_set',
                              'fence_set',
                              'out_set',
                              'va_obs_set',
                              'va_fence_set',
                              'va_out_set']

    def default_params(self):
        params = dict()
        params['min_epoch'] = 1
        params['max_epoch'] = 100
        params['batch_size'] = 128
        params['param_scale'] = 0.01
        params['patience'] = 25
        params['early_stop'] = False
        params['stop_criterion'] = 1e-3
        params['has_va_set'] = False
        params['gru_state_count'] = 50
        params['hmm_state_count'] = 5
        params['obs_lambda'] = 0
        params['l1_lambda'] = 0
        params['l2_lambda'] = 0
        params['obs_only'] = False

        if self.type in ['hmm', 'gru-hmm']:
            params['emission_type'] = 'bernoulli'

        if self.type in ['gru', 'gru-hmm']:
            params['max_conn'] = 0
            params['redun_epsilon'] = 0

        if self.type in ['gru-hmm-tree']:
            params['mlp_min_epoch'] = 1
            params['mlp_max_epoch'] = 1000

        return params

    def run(self, x, fcpt, y=None, va_x=None, va_fcpt=None, va_y=None):
        func = self.func
        keys = get_relevant_args(func, self.unused_params)
        used_params = dict((k, self.params[k]) for k in keys if k in self.params)
        if self.params['has_va_set']:
            used_params['va_obs_set'] = va_x
            used_params['va_fence_set'] = va_fcpt
            used_params['va_out_set'] = va_y

        pred_fun, loglike_fun, weights = func(x, fcpt, y, **used_params)

        self.params['pred_fun'] = pred_fun
        self.params['loglike_fun'] = loglike_fun
        self.params['trained_weights'] = weights

    def eval(self, tr_x, tr_fcpt, tr_y, te_x, te_fcpt, te_y):
        metrics = dict()

        trained_weights = self.params['trained_weights']
        pred_fun = self.params['pred_fun']
        loglike_fun = self.params['loglike_fun']

        if self.type in ['gru-hmm-tree']:
            trained_weights = trained_weights[0]
            pred_fun = pred_fun[0]
            loglike_fun = loglike_fun[0]

        # loglikelihoods
        if self.type in ['lr', 'hmm-sgd', 'gru-hmm']:
            metrics['{}:tr_ll_obs'.format(self.name)] = \
                loglike_fun(trained_weights, tr_x, tr_fcpt, None)
            metrics['{}:te_ll_obs'.format(self.name)] = \
                loglike_fun(trained_weights, te_x, te_fcpt, None)

        metrics['{}:tr_ll_out'.format(self.name)] = \
            loglike_fun(trained_weights, tr_x, tr_fcpt, tr_y)
        metrics['{}:te_ll_out'.format(self.name)] = \
            loglike_fun(trained_weights, te_x, te_fcpt, te_y)

        # aucs
        tr_pr_set = pred_fun(trained_weights, tr_x, tr_fcpt)
        te_pr_set = pred_fun(trained_weights, te_x, te_fcpt)

        tr_pr_set = tr_pr_set.reshape(tr_y.shape)
        te_pr_set = te_pr_set.reshape(te_y.shape)

        metrics['{}:tr_auc'.format(self.name)] = \
            get_auc(tr_y.flatten(), tr_pr_set.flatten())
        metrics['{}:te_auc'.format(self.name)] = \
            get_auc(te_y.flatten(), te_pr_set.flatten())

        # accuracy
        metrics['{}:tr_acc'.format(self.name)] = \
            get_accuracy(tr_y, tr_pr_set)
        metrics['{}:te_acc'.format(self.name)] = \
            get_accuracy(te_y, te_pr_set)

        # sparsity
        metrics['{}:sparsity'.format(self.name)] = \
            get_log_sparsity(trained_weights)

        pp_json(metrics)
        return metrics


class RLSTMPipeline(object):

    def __init__(self, dataset, log_file=None, weights_file=None):
        '''
            Args
            ----
            log_file : string
                       path to file to print training log to.
            weights_file : string
                           path to dir to save *npy and *pkl files.
        '''

        self.head = None
        self.tail = None
        self.num_nodes = 0

        self.log_file = log_file
        self.weights_file = weights_file
        self.dataset = dataset

        self.avail_models = ['lr',
                             'mlp',
                             'hmm-sgd',
                             'gru',
                             'gru-hmm',
                             'gru-hmm-tree']

        (tr_obs_set, tr_fcpt_set, tr_out_set), (te_obs_set, te_fcpt_set, te_out_set), \
            (va_obs_set, va_fcpt_set, va_out_set) = self.load_data(dataset)

        self.data = ((tr_obs_set, tr_fcpt_set, tr_out_set),
                     (te_obs_set, te_fcpt_set, te_out_set),
                     (va_obs_set, va_fcpt_set, va_out_set))
        self.obs_count = tr_obs_set.shape[0]
        self.out_count = tr_out_set.shape[0]

        print('[{}] initialized pipeline'.format(datetime.now()))

    def load_data(self, dataset):
        self.dataset = dataset
        tr_obs_set = np.load(os.path.join(dataset, 'X_train.npy'))
        tr_out_set = np.load(os.path.join(dataset, 'y_train.npy'))
        tr_fcpt_set = np.load(os.path.join(dataset, 'fcpt_train.npy'))
        te_obs_set = np.load(os.path.join(dataset, 'X_test.npy'))
        te_out_set = np.load(os.path.join(dataset, 'y_test.npy'))
        te_fcpt_set = np.load(os.path.join(dataset, 'fcpt_test.npy'))

        va_obs_set = None
        va_fcpt_set = None
        va_out_set = None

        if os.path.isfile(os.path.join(dataset, 'X_valid.npy')):
            va_obs_set = np.load(os.path.join(dataset, 'X_valid.npy'))
        if os.path.isfile(os.path.join(dataset, 'fcpt_valid.npy')):
            va_fcpt_set = np.load(os.path.join(dataset, 'fcpt_valid.npy'))
        if os.path.isfile(os.path.join(dataset, 'y_valid.npy')):
            va_out_set = np.load(os.path.join(dataset, 'y_valid.npy'))

        return (tr_obs_set, tr_fcpt_set, tr_out_set), \
            (te_obs_set, te_fcpt_set, te_out_set), \
            (va_obs_set, va_fcpt_set, va_out_set)

    def get_node(self, name=None, index=None):
        ''' Select a node (if exists) based on name or index

            Args
            ----
            name: string
                  name of node to select
            index: integer
                   index of node to select

            Returns
            -------
            node: RLSTMNode object
        '''
        node = self.head
        while node:
            if name and index:
                if node.name == name and node.index == index:
                    return node
            elif name:
                if node.name == name:
                    return node
            else:
                if node.index == index:
                    return node

            node = node.next
        return None

    def add_node(self, model_name, model_class, model_params=None):
        ''' Appends new model to end of file (or loc if specified)

            Args
            ----
            model_name: string
                        name of current node
            model_class: string
                        must be in 'avail_models'
            loc: integer
                 index of node. defaults last
        '''

        if not model_class in self.avail_models:
            return ValueError('model class {} not recognized.'.format(model_class))

        new_node = RLSTMNode(model_name,
                             model_class,
                             self.num_nodes,
                             None,
                             None,
                             model_params=model_params)

        if self.head is None:
            self.head = self.tail = new_node
        else:
            new_node.prev = self.tail
            new_node.next = None
            self.tail.next = new_node
            self.tail = new_node

        self.num_nodes += 1

        print('[{}] added new node {}. Total: {} nodes.'.format(
            datetime.now(), model_name, self.num_nodes))

    def run_and_eval(self):
        ''' Runs the entire pipeline. Each node is assumed
            to be a well built model, and is run in turn.
        '''

        print('[{}] executing pipeline'.format(datetime.now()))

        # grab the loaded data
        ((tr_obs_set, tr_fcpt_set, tr_out_set),
         (te_obs_set, te_fcpt_set, te_out_set),
         (va_obs_set, va_fcpt_set, va_out_set)) = self.data

        node = self.head
        while node:
            # check params for RLSTMShareWeights object
            # if so, replace w/ new weights
            for param in node.params:
                param_obj = node.params[param]
                if isinstance(param_obj, RLSTMShareWeights):
                    print('[{}] using parent ({}) weights'.format(
                        datetime.now(), param_obj.parent_name))
                    node.params[param] = param_obj.get_parent_weights(self)

            print('[{}] executing node ({})'.format(datetime.now(), node.name))
            node.run(tr_obs_set, tr_fcpt_set, tr_out_set, va_obs_set, va_fcpt_set, va_out_set)
            node = node.next

        node = self.head
        all_metrics = {}
        while node:
            node_metrics = node.eval(tr_obs_set,
                                     tr_fcpt_set,
                                     tr_out_set,
                                     te_obs_set,
                                     te_fcpt_set,
                                     te_out_set)

            all_metrics = merge_two_dicts(all_metrics, node_metrics)
            node = node.next

        if self.log_file:
            with open(self.log_file, 'w') as f:
                print(all_metrics, file=f)

    def save_weights(self):
        ''' to be used after run(). '''
        print('[{}] saving pipeline weights'.format(datetime.now()))

        w_dict = {}

        node = self.head
        while node:
            w_dict[node.name] = node.params['trained_weights']
            node = node.next

        if self.weights_file:
            cPickle.dump(w_dict, open(self.weights_file, 'w'))




from __future__ import print_function
from collections import defaultdict

import autograd.numpy as np
from autograd import value_and_grad, grad
from rlstm.optimize import adam
from rlstm.debug_utils import getCurMemCost_MiB


# -- train functions --


def _train_by_sgd(
        obs_set,
        loglike_fun,
        logpred_fun,
        num_weights,
        out_set=None,
        init_weights=None,
        min_epoch=100,
        max_epoch=500,
        batch_size=128,
        param_scale=0.01,
        # regularization
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        num_l1_weights=0,
        num_l2_weights=0,
        # optimization
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        regression=False,
        # validation set
        va_obs_set=None,
        va_out_set=None,
        # overwrite functions
        batch_indices=None,
        regularized_loss=None,
        batch_loss=None,
        callback=None):
    ''' Generic function for training using stochastic gradient descent.
        All models are wrapped around this function.
    '''

    # assume the *num_data* is always last index
    num_batches = int(np.ceil(obs_set.shape[-1] / float(batch_size)))

    if init_weights is None:
        init_weights = np.random.randn(num_weights) * param_scale

    # choose the proper error function
    error_func = get_mean_squared_error if regression else get_accuracy

    if batch_indices is None:
        def batch_indices(iter):
            idx = iter % num_batches
            return slice(idx * batch_size, (idx+1) * batch_size)

    # wrapper around loss function
    if regularized_loss is None:
        def regularized_loss(weights, obs_set,out_set):
            loss = loglike_fun(weights, obs_set, out_set)
            if not out_set is None and obs_lambda > 0:
                # Add a parameter to balance between out log-likelihood and obs log-likelihood
                loss += obs_lambda*loglike_fun(weights, obs_set, None)

            # negative log like + regularizers
            loss = -loss + \
                l1_lambda * np.sum(np.abs(weights[:num_l1_weights])) + \
                l2_lambda * np.sum(np.power(weights[:num_l2_weights], 2))
            return loss

    # loss function in batches
    if batch_loss is None:
        def batch_loss(weights, iter):
            idx = batch_indices(iter)
            batch_obs_set = obs_set[..., idx]
            batch_out_set = None if out_set is None else out_set[..., idx]
            return regularized_loss(weights, batch_obs_set, batch_out_set)

    if callback is None:
        def callback(e_i, b_i, ll, acc, va_ll, va_acc, x):
            print('epoch {:_<4}  |  batch {:_<2}/{:_<2}  |  loss {:_<10}  |  acc {:_<6}'.format(
                e_i,
                b_i+1,
                num_batches,
                round(ll, 4),
                round(acc, 4)))
            if b_i == num_batches - 1:
                print('======================================================================================================')
                print('EPOCH SUMMARY - LOSS {:_<10} | ACC {:_<6} | VA_LOSS {:_<10} | VA_ACC {:_<6} | MEM {}'.format(
                    round(ll, 4),
                    round(acc, 4),
                    round(va_ll, 4) if va_ll else 'nan',
                    round(va_acc, 4) if va_acc else 'nan',
                    int(getCurMemCost_MiB())))
                print('======================================================================================================')

    grad_fun = grad(batch_loss)
    weights = adam(grad_fun,
                   init_weights,
                   step_size=0.001,
                   num_batches=num_batches,
                   min_epoch=min_epoch,
                   max_epoch=max_epoch,
                   callback=callback,
                   stop_criterion=stop_criterion,
                   ll_fun=lambda x, i: batch_loss(x, i),
                   err_fun=lambda x, i: batch_error(x, i),
                   va_ll_fun=lambda x: regularized_loss(x, va_obs_set, va_out_set) if va_obs_set is not None else None,
                   va_err_fun=lambda x: error_func(va_out_set, np.exp(logpred_fun(x, va_obs_set))) if va_obs_set is not None else None,
                   early_stop=early_stop,
                   patience=patience)

    return weights



''' Often time series data may be too long for machines to
    handle. We need a good way of sub-sampling a long time
    series sequence.

    My plan is to for each data sequence, find the locations
    of the positive examples. Pick some number of them. And then
    subsample randomly for the negative examples bewteen the
    positive ones as well.
'''

import numpy as np


def subsample_array(A, num, return_indexes=False):
    indexes = np.sort(np.random.choice(xrange(A.size), size=num))
    if return_indexes:
        return indexes
    return A[indexes]


class LocalTimeSeriesSampler(object):
    ''' LocalTimeSeriesSampler

        Takes every single positive example but keeps only a
        local subset of the negative examples.

        Example:

        y = 0 0 0 0 0 1 0 0 0 0 0 0 0 0 1 0 1 0 1 0 0 1 0 0 0 0 0 0 1 0 0 0 0 1
        t = 1 2 3 4 5 6 7 8 9 a b c d e f g h i j k l m n o p q r s t u v w x y z

        First sequence: t = 1 2 3 4 5 6
        Second sequence: t = a b c d e f g h i j k l m
        Third sequence: t = o p q r s t u v w x y

        Args:
        pre_range <- number of local examples to take before every + example
             this should be large (50-500)
        post_range <- number of local examples to take after every + example
             this should be small (0-50)
        pos_frac <- fraction of positive examples to take (default 1)
    '''

    def _init__(self):
        pass

    def subsample(self, X, y, fcpt,
                              pre_range=100,
                              post_range=0,
                              pos_frac=1):
        ''' Args
            ----
            X : np array
               a (F x N) matrix where F is # features,
               N is a flattened time x data points
            y : np array
                is a (1 x N) matrix
                multidimensional outputs are not supported.
            fcpt : np array
                   a D + 1 array showing the splits between data
            pos_frac : double
            pre_range : integer
            post_range : integer

            Returns
            -------
            sub_X : np array
                    a (F x M) matrix where t is a subset of T
            sub_y : np array
                    a (O x M) matrix (binary)
            sub_fcpt : np array
                       a D'+1 array
        '''

        assert pos_frac >= 0 and pos_frac <= 1, \
            'fraction of positive examples must be between 0 and 1'
        assert pre_range >= 0, 'pre_range must be greater than 0'
        assert post_range >= 0, 'pre_range must be greater than 0'
        assert len(X.shape) == 2, '<X> must be 2 dimensional.'
        assert len(y.shape) == 2, '<y> must be 2 dimensional.'
        assert y.shape[0] == 1, '<y> can only have 1 dimension for now'
        assert list(np.unique(y)) == [0, 1], '<y> must only contain 0 or 1.'

        num_features, num_collapse = X.shape
        num_data = fcpt.size - 1

        sub_X, sub_y, sub_fcpt = [], [], [0]

        for data_i in range(num_data):
            fcpt_s_i = fcpt[data_i]
            fcpt_e_i = fcpt[data_i+1]
            num_time = fcpt_e_i - fcpt_s_i

            # grab the current outputs as a flat array
            cur_y = y[0, fcpt_s_i:fcpt_e_i]

            # find indexes of positive samples
            cur_y_where_pos = np.where(cur_y > 0)[0]
            if pos_frac < 1:
                cur_y_where_pos = np.random.choice(
                    cur_y_where_pos,
                    int(cur_y_where_pos.shape[0]*pos_frac))

            indexes_to_keep = []
            for i in range(len(cur_y_where_pos[:-1])):
                cur_idx = cur_y_where_pos[i]
                nex_idx = cur_y_where_pos[i+1]

                s_i = cur_idx-pre_range
                e_i = cur_idx+post_range
                if e_i >= nex_idx:
                    e_i = nex_idx - 1

                cur_slice = range(s_i, e_i+1)
                indexes_to_keep += cur_slice

            # do same for last index
            cur_idx = cur_y_where_pos[-1]
            s_i = cur_idx-pre_range
            e_i = num_time - 1
            cur_slice = range(s_i, e_i+1)
            indexes_to_keep += cur_slice
            indexes_to_keep = np.array(indexes_to_keep)
            indexes_to_keep = np.sort(np.unique(indexes_to_keep[indexes_to_keep >= 0]))
            # shift to true index
            indexes_to_keep += fcpt_s_i

            cur_X = X[:, indexes_to_keep]
            cur_y = y[:, indexes_to_keep]
            cur_fcpt = indexes_to_keep.size + fcpt_s_i

            sub_X.append(cur_X)
            sub_y.append(cur_y)
            sub_fcpt.append(cur_fcpt)

        sub_X = np.concatenate(sub_X, axis=1)
        sub_y = np.concatenate(sub_y, axis=1)
        sub_fcpt = np.array(sub_fcpt)

        return sub_X, sub_y, sub_fcpt



import os
import pdb
import copy
import shutil
import dill
import json
import cPickle
import numpy as np
from rlstm.pipeline import RLSTMNode, RLSTMPipeline
from rlstm.common_util import (is_json_serializable, is_numpy_array,
                               is_pickle_serializable, get_timestamp)

class Marmalade(object):
    ''' Like a Pickle or JSON but only for RLSTMNode and
        RLSTMPipeline objects
    '''

    @classmethod
    def serialize(cls, object, output_file):
        if isinstance(object, RLSTMPipeline):
            with open(output_file, 'w') as fp:
                cls.serialize_pipe(object, fp)
        elif isinstance(object, RLSTMNode):
            with open(output_file, 'w') as fp:
                cls.serialize_node(object, fp)
        else:
            with open(output_file, 'w') as fp:
                cPickle.dump(object, fp)

    @classmethod
    def serialize_pipe(cls, pipe, fp):
        ''' An entire pipeline can be saved as a large JSON object.
            Calling this function will serialize each node in turn

            The structure will look like the following:

            {
                'dataset': ...,
                'log_file': ...,
                'weights_file': ...,
                'num_nodes': ...,
                'node_0': {...},
                'node_1': {...},
                ...
            }
        '''
        output = {}
        output['dataset'] = pipe.dataset
        output['log_file'] = pipe.log_file
        output['weights_file'] = pipe.weights_file
        output['num_nodes'] = pipe.num_nodes

        index = 0
        node = pipe.head
        while node:
            node_json = cls.serialize_node(node, fp, return_string=True)
            output["node_{}".format(index)] = node_json
            node = node.next
            index += 1

        json.dump(output, fp,
                  sort_keys=True,
                  indent=4,
                  separators=(',', ': '))
        return

    @classmethod
    def unserialize_pipe(cls, fp):
        ''' deserializes a pipe and returns a RLSTMPipeline object '''
        input_json = json.load(fp)
        pipeline = RLSTMPipeline(input_json['dataset'],
                                 input_json['log_file'],
                                 input_json['weights_file'])
        num_nodes = input_json['num_nodes']

        for node_i in range(num_nodes):
            node = input_json['node_{}'.format(node_i)]
            node_params = cls._unserialize_node_params(node['params'])
            pipeline.add_node(node['name'],
                              node['type'],
                              model_params=node_params)

        return pipeline

    @classmethod
    def serialize_node(cls, node, fp, return_string=False):
        ''' Convert node into a json of the following format

            {
                'name': ...,
                'index': ...,
                'type': ...,
                'prev': ...,
                'next': ...,
                ...
                **kwparams,
            }

            - np.arrays will be stored as a path to a numpy serialization
            - Normal types will be stored retringgularly in JSON
            - All other types will be stored as a path to a pickle serialization
        '''

        # safely define/create a folder that all nodes will write to
        output_file = '{}_{}'.format('_marma_node_files', get_timestamp(True))
        intermediate_folder = os.path.join(os.path.dirname(fp.name), output_file)

        if not os.path.isdir(intermediate_folder):
            os.mkdir(intermediate_folder)

        node_json = {}
        node_json['name'] = node.name
        node_json['index'] = node.index
        node_json['type'] = node.type
        # node_json['prev'] = (node.prev.index, node.prev.name)
        # node_json['next'] = (node.next.index, node.next.name)

        params_json = {}
        for param in node.params:
            if is_json_serializable(node.params[param]):
                params_json[param] = node.params[param]
            elif is_numpy_array(node.params[param]):
                param_id = '_marmalade_{name}_{id}_{param}.npy'.format(
                    name=node.name, id=node.index, param=param)
                param_path = os.path.abspath(os.path.join(intermediate_folder, param_id))
                np.save(param_path, node.params[param])
                params_json[param] = param_path
            elif is_pickle_serializable(node.params[param]):
                param_id = '_marmalade_{name}_{id}_{param}.pkl'.format(
                    name=node.name, id=node.index, param=param)
                param_path = os.path.abspath(os.path.join(intermediate_folder, param_id))
                with open(param_path, 'w') as pfp:
                    cPickle.dump(node.params[param], pfp)
                params_json[param] = param_path
            else:
                param_id = '_marmalade_{name}_{id}_{param}.dill'.format(
                    name=node.name, id=node.index, param=param)
                param_path = os.path.abspath(os.path.join(intermediate_folder, param_id))
                with open(param_path, 'w') as pfp:
                    dill.dump(node.params[param], pfp)
                params_json[param] = param_path

        node_json['params'] = params_json
        if not return_string:
            json.dump(node_json, fp,
                      sort_keys=True,
                      indent=4,
                      separators=(',', ': '))
            return
        return node_json

    @classmethod
    def unserialize_node(cls, fp):
        ''' deserializes a node and returns a RLSTMNode object '''
        input_json = json.load(fp)
        node_params = cls._unserialize_node_params(input_json['params'])
        node = RLSTMNode(input_json['name'],
                         input_json['type'],
                         input_json['index'],
                         None,
                         None,
                         model_params=node_params)
        return node

    @classmethod
    def _unserialize_node_params(cls, params):
        ''' Load the numpy and cpickle files as needed '''
        for param in params:
            cur_param = str(params[param])
            if len(cur_param) > 4:
                if cur_param[-4:] == '.npy':
                    cur_param = np.load(cur_param)
                    params[param] = cur_param
                elif cur_param[-4:] == '.pkl':
                    with open(cur_param) as pfp:
                        cur_param = cPickle.load(pfp)
                        params[param] = cur_param
                elif cur_param[-5:] == '.dill':
                    with open(cur_param) as pfp:
                        cur_param = dill.load(pfp)
                        params[param] = cur_param
        return params



from __future__ import absolute_import, print_function

import pdb
from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.models import gru, hmm
from rlstm.weights_parser import WeightsParser
from rlstm.nn_util import sigmoid, add_bias_and_multiply
from rlstm.common_util import safe_log

# HMM provides continuous support for HMM through
# continuous emission functions.
from rlstm.emission_models import bernoulli_multivariate
from rlstm.emission_models import diagonal_gaussian_multivariate

# --------------------
#  GRUHMM CONSTRUCTOR
# --------------------

def build(
        input_count,
        hmm_state_count,
        gru_state_count,
        output_count,
        max_conn=0,
        redun_epsilon=0,
        emission_type='bernoulli'
    ):

    assert emission_type in ['bernoulli', 'gaussian'], 'EMISSION_TYPE not recognized.'
    if emission_type == 'bernoulli':
        emission_module = bernoulli_multivariate
    else:  # must be gaussian
        emission_module = diagonal_gaussian_multivariate
    calc_log_proba_arr_for_x = emission_module.calc_log_proba_data_arr_NK

    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)
    parser = WeightsParser()

    # GRU parameters
    parser.add_shape('init_hiddens', (1, gru_state_count))
    parser.add_shape('update_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('update_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('reset_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('reset_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('thidden_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('thidden_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('output_h_weights', (gru_state_count, output_count))

    num_gru_weights = parser.num_weights

    # HMM parameters
    parser.add_shape('lpi_mat', (hmm_state_count, 1))
    parser.add_shape('ltrans_mat', (hmm_state_count, hmm_state_count))
    parser.add_shape('eta_mat', (input_count, hmm_state_count))
    parser.add_shape('output_hmm_weights', (output_count, hmm_state_count + 1))

    num_hmm_weights = parser.num_weights - num_gru_weights

    # Make predictions through a linear combination of GRU and HMM parameters
    def outputs(
            weights,
            input_set,
            fence_set,
            output_set=None,
            return_lstate_set=False,
            return_pred_set=False
        ):

        # grab all gru weights
        update_x_weights = parser.get(weights, 'update_x_weights')
        update_h_weights = parser.get(weights, 'update_h_weights')
        reset_x_weights = parser.get(weights, 'reset_x_weights')
        reset_h_weights = parser.get(weights, 'reset_h_weights')
        thidden_x_weights = parser.get(weights, 'thidden_x_weights')
        thidden_h_weights = parser.get(weights, 'thidden_h_weights')
        output_h_weights = parser.get(weights, 'output_h_weights')

        # grab hmm weights
        lpi_mat = parser.get(weights, 'lpi_mat')
        ltrans_mat = parser.get(weights, 'ltrans_mat')
        eta_mat = parser.get(weights, 'eta_mat')
        output_hmm_weights = parser.get(weights, 'output_hmm_weights')

        # normalize probas
        lpi_mat = lpi_mat - logsumexp(lpi_mat, axis=0)
        ltrans_mat = ltrans_mat - logsumexp(ltrans_mat, axis=1, keepdims=True)
        start_log_proba_K = lpi_mat.flatten()

        data_count = len(fence_set) - 1
        feat_count = input_set.shape[0]

        hmm_ll = 0.0
        gru_ll = 0.0
        n_i_track = 0
        fence_base = fence_set[0]

        pred_set = None
        lstate_set = None

        if return_pred_set:
            pred_set = np.zeros((output_count, input_set.shape[1]))

        if return_lstate_set:
            lstate_set = np.zeros((hmm_state_count, input_set.shape[1]))

        # the output set of dimension ( output_count, time_count, data_count )
        # output_set = np.zeros( ( output_count, time_count, data_count ) )
        for data_iter in range(data_count):
            # Initialize hidden states in the HMM
            hiddens = copy(parser.get(weights, 'init_hiddens'))

            # use fenceposts to get time sequences
            fence_post_1 = fence_set[data_iter] - fence_base
            fence_post_2 = fence_set[data_iter+1] - fence_base
            time_count = fence_post_2 - fence_post_1

            # grab sequence for this data_iter
            curr_input_DT = input_set[:, fence_post_1:fence_post_2]

            # add bias and multiply weights into inputs all together
            weighted_input_HT = add_bias_and_multiply(update_x_weights, curr_input_DT.T).T

            # get log proba array from emissions
            curr_input_log_proba_TK = calc_log_proba_arr_for_x(curr_input_DT, eta_DK=eta_mat)

            # Initialize fwd belief vector at t = 0
            cur_belief_log_proba_K = start_log_proba_K + curr_input_log_proba_TK[0]
            cur_x_log_proba = logsumexp(cur_belief_log_proba_K)
            cur_belief_log_proba_K = cur_belief_log_proba_K - cur_x_log_proba

            # Loop over time steps
            for time_iter in range(time_count):
                # update gru
                hiddens = gru.update(
                    max_conn,
                    redun_epsilon,
                    np.expand_dims(weighted_input_HT[:, time_iter], axis=0),
                    hiddens,
                    update_x_weights,
                    update_h_weights,
                    reset_x_weights,
                    reset_h_weights,
                    thidden_x_weights,
                    thidden_h_weights
                )

                # update hmm
                if time_iter > 0:
                    cur_belief_log_proba_K, cur_x_log_proba = hmm.update_belief_log_probas(
                        cur_belief_log_proba_K,
                        curr_input_log_proba_TK[time_iter],
                        ltrans_mat
                    )

                #  get ll
                lstate_norm = np.expand_dims(cur_belief_log_proba_K, axis=1)
                curr_state = np.vstack((copy(lstate_norm), np.ones((1, lstate_norm.shape[1]))))

                if output_set is not None:
                    out_proba = sigmoid(np.sign(output_set[:, n_i_track] - 1e-3)
                                    * np.dot(hiddens, output_h_weights)
                                    + np.transpose(np.dot(output_hmm_weights, curr_state)))
                    out_lproba = safe_log(out_proba)
                    gru_ll += np.sum(out_lproba)
                else:
                    out_proba = sigmoid(np.dot(hiddens, output_h_weights)
                                    + np.transpose(np.dot(output_hmm_weights, curr_state)))
                    out_lproba = safe_log(out_proba)

                if return_pred_set:
                    pred_set[:, n_i_track] = out_lproba[0]

                if return_lstate_set:
                    lstate_set[:, n_i_track] = cur_belief_log_proba_K

                n_i_track += 1

            my_ll = logsumexp(cur_belief_log_proba_K)
            hmm_ll = hmm_ll + my_ll

        return gru_ll, pred_set, hmm_ll, lstate_set

    def prediction(weights, input_set, fence_set):
        _, out, _, _ = outputs(
            weights,
            input_set,
            fence_set,
            return_pred_set=True
        )
        return out

    def log_likelihood(weights, input_set, fence_set, output_set=None):
        gru_ll, _, hmm_ll, _ = outputs(
            weights,
            input_set,
            fence_set,
            output_set=output_set,
            return_pred_set=False,
            return_lstate_set=False
        )

        return hmm_ll if output_set is None else gru_ll

    return prediction, log_likelihood, num_gru_weights, num_hmm_weights



'''
Example containing a simple 2D distribution of sampled
points along a uniform circle (4000 training points).

We will pass this into the MUNGE function to expand to 
40,000 training points. The resulting sample distribution
should closely resemble that of the uniform circle.
'''

import numpy as np
import matplotlib.pyplot as plt

# make the toy data
r = 0.5
num = 10000
x = np.random.uniform( low=0 , high=1 , size=num ) - 0.5
y = np.sqrt( r**2 - x**2 )
y[num/2:] *= -1

# add some noise 
param_scale = 0.25
x_noise = ( np.random.random( num ) - 0.5 ) * param_scale
y_noise = ( np.random.random( num ) - 0.5 ) * param_scale
x += x_noise
y += y_noise

# plt.figure()
# plt.plot(x, y, 'o')
# plt.savefig('./originaldata.png')
# plt.show()

T = np.vstack( ( x, y ) ).T

# munge the data
import munge
k = 10 # make 100k points
p = 0.5
s = 1.0
A_type = [True, True]

T_munged = munge.munge( T , k , p , s , A_type=A_type )

x , y = T_munged[:, 0], T_munged[:, 1]

plt.figure()
plt.plot(x, y, 'o')
plt.savefig('./mungeddata.png')
plt.show()


import pdb
import autograd.numpy as np
from autograd.scipy.misc import logsumexp
from rlstm.models.hmm import *


if __name__ == '__main__':
    from rlstm.emission_models import bernoulli_multivariate as emission_module

    # Load simple toy dataset
    dataset_path = os.path.join('../../datasets/tiny_obs_and_out/')
    x_arr = np.load(os.path.join(dataset_path, 'obs_set.npy'))

    n_dims, n_timesteps, n_seqs = x_arr.shape
    x_DM = x_arr.swapaxes(0, 2).reshape((-1, n_dims)).T
    fenceposts_Np1 = np.arange(0, (n_seqs + 1) * n_timesteps, n_timesteps)

    # TEST #1
    # Make a parameter set with just one state
    # Verify that loss calculation for full hmm
    # is exactly equal to just using the emission model to calc log probas
    K = 1
    calc_log_proba_arr_for_x = emission_module.calc_log_proba_data_arr_NK
    eta_DK = emission_module.init_natural_param_array_from_data(
        data_DN=x_DM,
        K=K)
    start_log_proba_K = np.log(np.ones(K) / float(K))
    trans_log_proba_KK = np.ones((K,K))
    trans_log_proba_KK /= trans_log_proba_KK.sum(axis=1)[:,np.newaxis]
    trans_log_proba_KK = np.log(trans_log_proba_KK)

    # Compute loss for single seq
    loss_x, loss_y, _ = calc_loss_for_one_sequence(
        x_n_DT=x_DM,
        calc_log_proba_arr_for_x=calc_log_proba_arr_for_x,
        eta_DK=eta_DK,
        start_log_proba_K=start_log_proba_K,
        trans_log_proba_KK=trans_log_proba_KK)

    x_log_proba_T = calc_log_proba_arr_for_x(x_DM, eta_DK=eta_DK)
    assert np.allclose(loss_x, np.sum(x_log_proba_T))
    print('>> PASSED TEST #1')

    # TEST #2
    # Make a parameter set with K=5 states
    # with trans probabilities that are UNIFORM from every state
    # Verify that loss calculation for full hmm
    # is exactly equal to just using a mixture model
    K = 5
    calc_log_proba_arr_for_x = emission_module.calc_log_proba_data_arr_NK
    eta_DK = emission_module.init_natural_param_array_from_data(
        data_DN=x_DM,
        K=K)
    start_log_proba_K = np.log(np.ones(K) / float(K))
    trans_log_proba_KK = np.ones((K,K))
    trans_log_proba_KK /= trans_log_proba_KK.sum(axis=1)[:,np.newaxis]
    trans_log_proba_KK = np.log(trans_log_proba_KK)

    # Compute loss for single seq
    hmm_loss_x, _, _ = calc_loss_for_one_sequence(
        x_n_DT=x_DM,
        calc_log_proba_arr_for_x=calc_log_proba_arr_for_x,
        eta_DK=eta_DK,
        start_log_proba_K=start_log_proba_K,
        trans_log_proba_KK=trans_log_proba_KK)

    x_log_proba_TK = calc_log_proba_arr_for_x(x_DM, eta_DK=eta_DK)
    prior_log_proba_K = start_log_proba_K
    mixture_loss_x = -1 * np.sum(logsumexp(
        x_log_proba_TK + prior_log_proba_K[np.newaxis, :], axis=1))
    assert np.allclose(hmm_loss_x, -mixture_loss_x)
    print('>> PASSED TEST #2')

    # TEST #3
    # Make a parameter set with K=5 states
    # with trans probabilities that are same but not uniform from every state
    # Verify that loss calculation for full hmm
    # is exactly equal to just using a mixture model
    K = 4
    calc_log_proba_arr_for_x = emission_module.calc_log_proba_data_arr_NK
    eta_DK = emission_module.init_natural_param_array_from_data(
        data_DN=x_DM,
        K=K,
        seed=123)
    start_log_proba_K = np.log(np.asarray([0.1, 0.2, 0.3, 0.4]))
    trans_log_proba_KK = np.tile(start_log_proba_K, (4,1))

    # Compute loss for single seq
    hmm_loss_x, _, _ = calc_loss_for_one_sequence(
        x_n_DT=x_DM,
        calc_log_proba_arr_for_x=calc_log_proba_arr_for_x,
        eta_DK=eta_DK,
        start_log_proba_K=start_log_proba_K,
        trans_log_proba_KK=trans_log_proba_KK)

    # Compute loss for same data under mixture model
    x_log_proba_TK = calc_log_proba_arr_for_x(x_DM, eta_DK)
    prior_log_proba_K = start_log_proba_K
    mixture_loss_x = -1 * np.sum(logsumexp(
        x_log_proba_TK + prior_log_proba_K[np.newaxis, :], axis=1))
    assert np.allclose(hmm_loss_x, -mixture_loss_x)
    print('>> PASSED TEST #3')

    # Finally, compute loss over many sequences
    hmm_loss_x, _, _ = calc_loss_for_many_sequences(
        x_DM=x_DM,
        fenceposts_Np1=fenceposts_Np1,
        calc_log_proba_arr_for_x=calc_log_proba_arr_for_x,
        eta_DK=eta_DK,
        start_log_proba_K=start_log_proba_K,
        trans_log_proba_KK=trans_log_proba_KK)
    print('ALL SEQ LOSS: {}'.format(hmm_loss_x))
    assert np.allclose(hmm_loss_x, -1933.99577891)



import autograd.numpy as np

from autograd import grad
from autograd.util import flatten
from scipy.optimize import fmin_l_bfgs_b

from bernoulli_multivariate import (
    init_natural_param_dict_from_data,
    calc_log_proba_data_arr_NK,
    mu2eta,
    eta2mu,
)

def make_problem(n_examples=1000, seed=0):
    prng = np.random.RandomState(seed)
    mu_DK = np.asarray([[0.1, 0.2, 0.3, 0.4, 0.5]]).T
    data_DN = prng.rand(mu_DK.shape[0], n_examples) <= mu_DK
    return data_DN, mu_DK

if __name__ == '__main__':
    np.set_printoptions(precision=3, suppress=1)

    data_DN, true_mu_DK = make_problem(n_examples=100000, seed=0)
    D, K = true_mu_DK.shape

    param_dict = init_natural_param_dict_from_data(data_DN)
    init_flat_vec, unflatten_func = flatten(param_dict)

    def calc_neg_log_proba_dataset_wrt_flat_eta(flat_vector):
        param_dict = unflatten_func(flat_vector)
        return -1 * np.sum(
            calc_log_proba_data_arr_NK(data_DN, **param_dict))
    grad_of_neg_log_proba_dataset_wrt_flat_eta = \
        grad(calc_neg_log_proba_dataset_wrt_flat_eta)

    fromrand_flat_vec, fromrand_loss_val, fromrand_dict = fmin_l_bfgs_b(
        calc_neg_log_proba_dataset_wrt_flat_eta,
        init_flat_vec,
        fprime=grad_of_neg_log_proba_dataset_wrt_flat_eta)
    fromrand_mu_DK = eta2mu(**unflatten_func(fromrand_flat_vec))

    fromtrue_flat_vec, fromtrue_loss_val, fromtrue_dict = fmin_l_bfgs_b(
        calc_neg_log_proba_dataset_wrt_flat_eta,
        mu2eta(true_mu_DK).flatten(),
        fprime=grad_of_neg_log_proba_dataset_wrt_flat_eta)
    fromtrue_mu_DK = eta2mu(**unflatten_func(fromtrue_flat_vec))

    print "TRUE MU"
    print ' '.join(['%.3f' % a for a in true_mu_DK.flatten()])
    print "\n OPTIMAL MU FROM RAND INIT"
    print ' '.join(['%.3f' % a for a in fromrand_mu_DK.flatten()])
    print "\n OPTIMAL MU FROM TRUTH"
    print ' '.join(['%.3f' % a for a in fromtrue_mu_DK.flatten()])



'''
Gaussian Multivariate observation model, with unknown mean and diag covariance.

Parameters
----------
K : num states
D : num dimensions for each observation

mu_DK : mean position for cluster k at dimension d
stddev_DK : standard deviation for cluster k at dimension d
'''

import autograd.numpy as np
from autograd.util import flatten
import sys
import pdb
import os

from autograd.scipy.misc import logsumexp

def init_natural_param_dict_from_data(
        data_DN, K=1, n_states=None, seed=0):
    '''
    '''
    prng = np.random.RandomState(int(seed))
    if n_states is not None:
        K = int(n_states)
    else:
        K = int(K)
    D = data_DN.shape[0]
    mu_DK = prng.randn(K, D).T
    stddev_DK = 0.1 + 0.1 * prng.rand(D, K)
    return mu2eta(
        mu_DK=mu_DK,
        stddev_DK=stddev_DK)


def init_natural_param_array(D, K, seed=0):
    prng = np.random.RandomState(int(seed))
    D, K = int(D), int(K)
    mu_DK = prng.randn(K, D).T
    stddev_DK = 0.1 + 0.1 * prng.rand(D, K)
    eta = mu2eta(mu_DK=mu_DK, stddev_DK=stddev_DK)
    return eta['eta_mu_DK'], eta['eta_stddev_DK']


def eta2mu(eta_mu_DK=None, eta_stddev_DK=None, eps=0.00001, **kwargs):
    ''' Convert unconstrained to mean parameters

    Returns
    -------
    mu_DK : 2D array, D x K
    stddev_DK : 2D array, D x K
    '''
    mu_DK = eta_mu_DK
    stddev_DK = eps + np.log(1.0 + np.exp(eta_stddev_DK))
    return dict(mu_DK=mu_DK, stddev_DK=stddev_DK)


def mu2eta(mu_DK=None, stddev_DK=None, eps=0.00001, **kwargs):
    ''' Convert mean to unconstrained parameters

    Returns
    -------
    eta_mu_DK : 2D array, D x K
    eta_stddev_DK : 2D array, D x K

    Examples
    --------
    >>> np.set_printoptions(suppress=1, precision=3)
    >>> mu_DK = np.zeros((4, 1))
    >>> stddev_DK = np.asarray([[0.1, 0.5, 1.0, 9.0]]).T
    >>> eta_dict = mu2eta(mu_DK=mu_DK, stddev_DK=stddev_DK)
    >>> print eta_dict['eta_mu_DK'].T
    [[ 0.  0.  0.  0.]]
    >>> print eta_dict['eta_stddev_DK'].T
    [[-2.252 -0.433  0.541  9.   ]]

    # Verify invertibility. If we transform then invert, we get original input.
    >>> mu_dict = eta2mu(**eta_dict)
    >>> np.allclose(mu_dict['mu_DK'], mu_DK)
    True
    >>> np.allclose(mu_dict['stddev_DK'], stddev_DK)
    True
    '''
    eta_mu_DK = mu_DK
    stddev_DK = np.maximum(stddev_DK, eps)
    eta_stddev_DK = np.log(np.exp(stddev_DK - eps) - 1.0)
    return dict(eta_mu_DK=eta_mu_DK, eta_stddev_DK=eta_stddev_DK)

def calc_log_proba_data_arr_NK(
        data_DN,
        mu_DK=None,
        stddev_DK=None,
        eta_mu_DK=None,
        eta_stddev_DK=None,
        **kwargs):
    ''' Compute log probability of each dataset observation under each cluster

    Args
    ----
    data_DN : 2D array, n_dims x n_examples, with real values
        data_DN[d, n] = observed value for observation n's d-th dimension
    mu_DK : 2D array, n_dims x n_states, with real values

    Returns
    -------
    log_proba_NK : 2D array, n_examples x n_states, with real values
        entry n,k gives log probability of data row n under state k

    Examples
    --------
    >>> np.set_printoptions(suppress=0, precision=3)
    >>> prng = np.random.RandomState(0)

    # Verify the log proba calc is correct (integrates to unity)
    >>> data_DN = np.linspace(-5, 5, 100)[:,np.newaxis].T
    >>> mu_DK = np.asarray([[0.0, 1.0]])
    >>> stddev_DK = np.asarray([[1.0, 0.5]])
    >>> log_proba_NK = calc_log_proba_data_arr_NK(
    ...     data_DN, mu_DK=mu_DK, stddev_DK=stddev_DK)
    >>> assert log_proba_NK.ndim == 2
    >>> sum_proba_1 = np.trapz(np.exp(log_proba_NK[:,0]), data_DN[0, :])
    >>> np.allclose(1.0, sum_proba_1)
    True
    >>> sum_proba_2 = np.trapz(np.exp(log_proba_NK[:,1]), data_DN[0, :])
    >>> np.allclose(1.0, sum_proba_2)
    True
    '''
    N = data_DN.shape[1]
    if eta_mu_DK is not None:
        mu_dict = eta2mu(eta_mu_DK=eta_mu_DK, eta_stddev_DK=eta_stddev_DK)
        mu_DK = mu_dict['mu_DK']
        stddev_DK = mu_dict['stddev_DK']

    D, K = mu_DK.shape
    log_proba_list = list()
    for k in xrange(K):
        # pdb.set_trace()
        var_k_D = np.square(stddev_DK[:, k])
        arr_k_ND = data_DN.T - mu_DK[:, k]
        arr_k_ND = np.square(arr_k_ND)
        arr_k_ND = arr_k_ND / var_k_D
        mahal_dist_k_N = np.sum(arr_k_ND, axis=1)
        log_proba_list.append(
            - 0.5 * D * np.log(2 * np.pi)
            - 0.5 * mahal_dist_k_N
            - np.sum(np.log(stddev_DK[:, k])))
    log_proba_NK = np.vstack(log_proba_list).T
    return log_proba_NK


# -- wrapper functions to handle single param vector --

def make_eta_D2K(eta_mu_DK, eta_stddev_DK):
    return np.hstack((eta_mu_DK, eta_stddev_DK))


def split_eta_D2K(eta_D2K):
    D, K2 = eta_D2K.shape
    K = int(K2 / 2)
    return eta_D2K[:, :K], eta_D2K[:, K:]


def calc_log_proba_data_arr_NK_flat(data_DN, eta_DK=None):
    ''' wrapper around calc_log_proba_data_arr_NK function that accepts
        a single array with the concatenated eta_mu_DK and eta_stdev_DK
    '''

    if eta_DK is not None:
        eta_mu_DK, eta_stddev_DK = split_eta_D2K(eta_DK)

    log_proba_NK = calc_log_proba_data_arr_NK(
        data_DN,
        eta_mu_DK=eta_mu_DK,
        eta_stddev_DK=eta_stddev_DK)

    return log_proba_NK



import argh
from argh import arg
import cPickle
from os.path import join
import autograd.numpy as np
import common_util
import pdb


def normalize(X, axis=0):
    return np.divide(X.T, np.sum(X, axis=axis)).T


def toy_matrices():
    pi_mat = np.array([.5, .5, 0, 0, 0])
    trans_mat = np.array(([.5, .5, 0, 0, 0],
                          [.25, .5, .25, 0, 0],
                          [0, .25, .5, .25, 0],
                          [0, 0, .25, .5, .25],
                          [0, 0, 0, .5, .5]))
    obs_mat = np.array(([.5, .5, .5, 0, 0, 0, 0],
                        [0, .5, .5, .5, 0, 0, 0],
                        [0, 0, .5, .5, .5, 0, 0],
                        [0, 0, 0, .5, .5, .5, 0],
                        [0, 0, 0, 0, .5, .5, .5]))
    # obs_mat = np.array( ( [ 1, 0, 0, 0, 0, 0, 0 ], \
    #                       [ 0, 1, 0, 0, 0, 0, 0 ], \
    #                       [ 0, 0, 1, 0, 0, 0, 0 ], \
    #                       [ 0, 0, 0, 1, 0, 0, 0 ], \
    #                       [ 0, 0, 0, 0, 1, 0, 0 ] ) )
    return pi_mat, trans_mat, obs_mat


def spurious_matrices():
    pi_mat = np.array([.2, .2, .2, .2, .2])
    trans_mat = np.array(([.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2],
                          [.2, .2, .2, .2, .2]))
    obs_mat = np.array(([.5, .5, .5, 0, 0, 0, 0],
                        [0, .5, .5, .5, 0, 0, 0],
                        [0, 0, .5, .5, .5, 0, 0],
                        [0, 0, 0, .5, .5, .5, 0],
                        [0, 0, 0, 0, .5, .5, .5]))
    return pi_mat, trans_mat, obs_mat


def toy_gaussian_matrices():
    mu_mat = np.array(([0, 0, 0, 0, 0, 0, 0],
                       [1, 1, 1, 1, 1, 1, 1],
                       [2, 2, 2, 2, 2, 2, 2],
                       [3, 3, 3, 3, 3, 3, 3],
                       [4, 4, 4, 4, 4, 4, 4]))
    sigma_mat = np.ones((5, 7))*np.sqrt(0.1)
    return mu_mat, sigma_mat


def spurious_gaussian_matrices():
    mu_mat = np.zeros((5, 7))
    sigma_mat = np.ones((5, 7))
    return mu_mat, sigma_mat


def create_discrete_toy_data(
        data_count, time_count, bias_mat=np.array([15, 15]),
        weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                             [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):

    # N is the number of samples and T is the length of the series.
    # Store data in DxTxN array because we'll want to slice by samples.

    # weight_mat is Ox(K+D)-dimensional.
    # bias_mat is O-dimensional (determines the proportion of 0s and 1s in
    # binary output).

    # Define the HMM:
    #   K is the number of states
    #   trans_mat(k,k') is the probability that we transition to k' from k (KxK-dimensional)
    # obs_mat(k,d) is the probability of 1 in dimension d in state k
    # (KxD-dimensional)
    state_count = 5  # number of states K (default = 5)
    # observation dimension D (default=7)
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]  # output dimension O (default=2)

    pi_mat, trans_mat, obs_mat = toy_matrices()

    # Create the sequences
    obs_set = np.zeros((dim_count, time_count, data_count), dtype=int)
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set = np.zeros((state_count, time_count, data_count), dtype=int)
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state = np.random.multinomial(1, pi_mat)
                state_set[:, 0, data_iter] = state
            else:
                tvec = np.dot(state_set[:, time_iter - 1, data_iter], trans_mat)
                state = np.random.multinomial(1, tvec)
                state_set[:, time_iter, data_iter] = state
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            obs_vec = np.dot(state_set[:, time_iter, data_iter], obs_mat)
            obs = np.random.binomial(1, obs_vec)
            obs_set[:, time_iter, data_iter] = obs
            in_vec = np.hstack((state_set[:, time_iter, data_iter],
                                obs_set[:, time_iter, data_iter]))
            out_vec = 1 / (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    # collect the outputs
    param_set = {'state_set': state_set,
                 'pi_mat': pi_mat,
                 'trans_mat': trans_mat,
                 'obs_mat': obs_mat,
                 'weight_mat': weight_mat,
                 'bias_mat': bias_mat}

    # output
    return obs_set, out_set, param_set


def create_continuous_toy_data(
        data_count, time_count, bias_mat=np.array([15, 15]),
        weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                             [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):

    ''' Similar to the discrete toy data but sampling from Gaussian
        to create a continuous state based dataset.
    '''
    state_count = 5
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]

    pi_mat, trans_mat, _ = toy_matrices()  # obs_mat not useful here
    mu_mat, sigma_mat = toy_gaussian_matrices()

    obs_set = np.zeros((dim_count, time_count, data_count))
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set = np.zeros((state_count, time_count, data_count), dtype=int)

    # generate state sequences
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state = np.random.multinomial(1, pi_mat)
                state_set[:, 0, data_iter] = state
            else:
                tvec = np.dot(state_set[:, time_iter - 1, data_iter], trans_mat)
                state = np.random.multinomial(1, tvec)
                state_set[:, time_iter, data_iter] = state

    # generate observations
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            state = state_set[:, time_iter, data_iter]
            state_i = np.argmax(state, axis=0)
            mu_set = mu_mat[state_i, :]
            sigma_set = sigma_mat[state_i, :]
            obs = np.random.randn(dim_count)*sigma_set + mu_set
            obs_set[:, time_iter, data_iter] = obs

            # generate outputs
            in_vec = np.hstack((state, obs))
            out_vec = 1 / (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    param_set = {'state_set': state_set,
                 'pi_mat': pi_mat,
                 'trans_mat': trans_mat,
                 'mu_mat': mu_mat,
                 'sigma_mat': sigma_mat,
                 'weight_mat': weight_mat,
                 'bias_mat': bias_mat}

    return obs_set, out_set, param_set


def create_discrete_2hmm_toy_data(
        data_count, time_count, bias_mat=np.array([15, 15]),
        weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                             [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):

    # Define two independent HMMs. Both will help generate the observations but one
    # one of them will generate the output. The HMM not modelling output is
    # basically noise.

    state_count = 5  # number of states K (default = 5)
    # observation dimension D (default=7)
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]  # output dimension O (default=2)

    pi_mat_1, trans_mat_1, obs_mat_1 = toy_matrices()  # hmm1
    pi_mat_2, trans_mat_2, obs_mat_2 = spurious_matrices()  # hmm2

    # store results here (int because only integers)
    obs_set = np.zeros((dim_count * 2, time_count, data_count), dtype=int)
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set_1 = np.zeros((state_count, time_count, data_count), dtype=int)  # hmm1
    state_set_2 = np.zeros((state_count, time_count, data_count), dtype=int)  # hmm2

    # create state sequences for both
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state_1 = np.random.multinomial(1, pi_mat_1)
                state_2 = np.random.multinomial(1, pi_mat_2)
                state_set_1[:, 0, data_iter] = state_1
                state_set_2[:, 0, data_iter] = state_2
            else:
                tvec_1 = np.dot(state_set_1[:, time_iter - 1, data_iter], trans_mat_1)
                tvec_2 = np.dot(state_set_2[:, time_iter - 1, data_iter], trans_mat_2)
                state_1 = np.random.multinomial(1, tvec_1)
                state_2 = np.random.multinomial(1, tvec_2)
                state_set_1[:, time_iter, data_iter] = state_1
                state_set_2[:, time_iter, data_iter] = state_2

    # generate obs and out matrices
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            # obs
            obs_vec_1 = np.dot(state_set_1[:, time_iter, data_iter], obs_mat_1)
            obs_vec_2 = np.dot(state_set_2[:, time_iter, data_iter], obs_mat_2)
            obs_1 = np.random.binomial(1, obs_vec_1)
            obs_2 = np.random.binomial(1, obs_vec_2)
            obs = np.hstack((obs_1, obs_2))  # concat together
            obs_set[:, time_iter, data_iter] = obs

            # out
            in_vec = np.hstack((state_set_1[:, time_iter, data_iter],  # only use hmm1 to make output
                                obs_set[:, time_iter, data_iter][:dim_count]))
            out_vec = 1 / (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    # collect the outputs
    param_set = {
        'state_set_1': state_set_1,
        'pi_mat_1': pi_mat_1,
        'trans_mat_1': trans_mat_1,
        'obs_mat_1': obs_mat_1,
        'state_set_2': state_set_2,
        'pi_mat_2': pi_mat_2,
        'trans_mat_2': trans_mat_2,
        'obs_mat_2': obs_mat_2,
        'weight_mat': weight_mat,
        'bias_mat': bias_mat
    }

    return obs_set, out_set, param_set


def create_continuous_2hmm_toy_data(
        data_count, time_count, bias_mat=np.array([15, 15]),
        weight_mat=np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                             [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))):
    state_count = 5
    dim_count = np.shape(weight_mat)[1] - state_count
    out_count = np.shape(weight_mat)[0]

    pi_mat_1, trans_mat_1, _ = toy_matrices()
    pi_mat_2, trans_mat_2, _ = spurious_matrices()
    mu_mat_1, sigma_mat_1 = toy_gaussian_matrices()
    mu_mat_2, sigma_mat_2 = spurious_gaussian_matrices()

    obs_set = np.zeros((dim_count * 2, time_count, data_count))
    out_set = np.zeros((out_count, time_count, data_count), dtype=int)
    state_set_1 = np.zeros((state_count, time_count, data_count), dtype=int)  # hmm1
    state_set_2 = np.zeros((state_count, time_count, data_count), dtype=int)  # hmm2

    # create state sequences for both
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            if time_iter == 0:
                state_1 = np.random.multinomial(1, pi_mat_1)
                state_2 = np.random.multinomial(1, pi_mat_2)
                state_set_1[:, 0, data_iter] = state_1
                state_set_2[:, 0, data_iter] = state_2
            else:
                tvec_1 = np.dot(state_set_1[:, time_iter - 1, data_iter], trans_mat_1)
                tvec_2 = np.dot(state_set_2[:, time_iter - 1, data_iter], trans_mat_2)
                state_1 = np.random.multinomial(1, tvec_1)
                state_2 = np.random.multinomial(1, tvec_2)
                state_set_1[:, time_iter, data_iter] = state_1
                state_set_2[:, time_iter, data_iter] = state_2

    # generate observations
    for data_iter in range(data_count):
        for time_iter in range(time_count):
            state_1 = state_set_1[:, time_iter, data_iter]
            state_2 = state_set_2[:, time_iter, data_iter]
            state_1_i = np.argmax(state_1, axis=0)
            state_2_i = np.argmax(state_2, axis=0)

            mu_set_1 = mu_mat_1[state_1_i, :]
            sigma_set_1 = sigma_mat_1[state_1_i, :]
            mu_set_2 = mu_mat_2[state_2_i, :]
            sigma_set_2 = sigma_mat_2[state_2_i, :]

            obs_1 = np.random.randn(dim_count)*sigma_set_1 + mu_set_1
            obs_2 = np.random.randn(dim_count)*sigma_set_2 + mu_set_2
            obs = np.hstack((obs_1, obs_2))
            obs_set[:, time_iter, data_iter] = obs

            # generate outputs
            in_vec = np.hstack((state_1, obs[:dim_count]))
            out_vec = 1 / (1 + np.exp(-1 * (np.dot(weight_mat, in_vec) - bias_mat)))
            out = np.random.binomial(1, out_vec)
            out_set[:, time_iter, data_iter] = out

    param_set = {
        'state_set_1': state_set_1,
        'pi_mat_1': pi_mat_1,
        'trans_mat_1': trans_mat_1,
        'state_set_2': state_set_2,
        'pi_mat_2': pi_mat_2,
        'trans_mat_2': trans_mat_2,
        'mu_mat_1':mu_mat_1,
        'mu_mat_2':mu_mat_2,
        'sigma_mat_1':sigma_mat_1,
        'sigma_mat_2':sigma_mat_2,
        'weight_mat': weight_mat,
        'bias_mat': bias_mat
    }

    return obs_set, out_set, param_set


@arg('data_type', help='select from toy_obs_only, toy_obs_and_out, toy_two_hmm')
@arg('-tr', '--train_count', default=20, help='number of data groups in train set')
@arg('-te', '--test_count', default=10, help='number of data groups in test set')
@arg('-va', '--valid_count', default=0, help='number of data groups in validation set')
@arg('-t', '--time_count', default=50, help='number of time indices per group')
@arg('-s', '--save_path', default='.', help='directory to save numpy files')
@arg('-b', '--binary', default=1, help='whether to generated binary or continuous data')
def output_toy_data(data_type,
                    train_count=20,
                    test_count=10,
                    valid_count=0,
                    time_count=50,
                    save_path=None,
                    binary=1):
    ''' Runs create_discrete_toy_data or create_discrete_2hmm_toy_data and saves
        it into the X_train, y_train, ... style

        Args
        ----
        data_type : string
                    Choose from toy_obs_only, toy_obs_and_out, toy_two_hmm
        train_count : int
                      Number of groups in training set
        test_count : int
                     Number of groups in testing set
        valid_count : int
                      Number of groups in validation set
        time_count : int
                     Number of time steps per group
        save_path : string / None
                    Where to save the outputs
        binary : boolean
                 whether to generated binary or continuous data

        Returns
        -------
        X_train : np array
        y_train : np array
        X_test : np array
        y_test : np array
        X_valid : np array
        y_valid : np array

    '''
    binary = bool(binary)
    data_count = train_count + test_count + valid_count

    if data_type == 'toy_obs_only':
        bias_mat = np.array([5, 5])
        weight_mat = np.array(([10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
        func = create_discrete_toy_data if binary else create_continuous_toy_data
        obs_set, out_set, param_set = func(data_count, time_count, bias_mat, weight_mat)
    elif data_type == 'toy_obs_and_out':
        bias_mat = np.array([15, 15])
        weight_mat = np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))
        func = create_discrete_toy_data if binary else create_continuous_toy_data
        obs_set, out_set, param_set = func(data_count, time_count, bias_mat, weight_mat)
    elif data_type == 'toy_two_hmm':
        bias_mat = np.array([15, 15])
        weight_mat = np.array(([10, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0],
                               [0, 10, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0]))
        func = create_discrete_2hmm_toy_data if binary else create_continuous_2hmm_toy_data
        obs_set, out_set, param_set = func(data_count, time_count, bias_mat, weight_mat)

    if save_path is None:
        save_path = join('datasets', data_type)

    np.save(join(save_path, 'obs_set.npy'), obs_set)
    np.save(join(save_path, 'out_set.npy'), out_set)
    with open(join(save_path, 'param_set.pkl'), 'wb') as fp:
        cPickle.dump(param_set, fp)

    # do a train_test_split by hand
    train_obs_set = obs_set[:, :, :train_count]
    train_out_set = out_set[:, :, :train_count]
    test_obs_set = obs_set[:, :, train_count:train_count+test_count]
    test_out_set = out_set[:, :, train_count:train_count+test_count]
    valid_obs_set = obs_set[:, :, train_count+test_count:data_count]
    valid_out_set = out_set[:, :, train_count+test_count:data_count]

    train_obs_set, train_fcpt_set = common_util.map_3d_to_2d(train_obs_set)
    valid_obs_set, valid_fcpt_set = common_util.map_3d_to_2d(valid_obs_set)
    test_obs_set, test_fcpt_set = common_util.map_3d_to_2d(test_obs_set)

    train_out_set, _ = common_util.map_3d_to_2d(train_out_set)
    valid_out_set, _ = common_util.map_3d_to_2d(valid_out_set)
    test_out_set, _ = common_util.map_3d_to_2d(test_out_set)

    np.save(join(save_path, 'X_train.npy'), train_obs_set)
    np.save(join(save_path, 'y_train.npy'), train_out_set)
    np.save(join(save_path, 'fcpt_train.npy'), train_fcpt_set)
    np.save(join(save_path, 'X_test.npy'), test_obs_set)
    np.save(join(save_path, 'y_test.npy'), test_out_set)
    np.save(join(save_path, 'fcpt_test.npy'), test_fcpt_set)
    np.save(join(save_path, 'X_valid.npy'), valid_obs_set)
    np.save(join(save_path, 'y_valid.npy'), valid_out_set)
    np.save(join(save_path, 'fcpt_valid.npy'), valid_fcpt_set)

if __name__ == '__main__':
    argh.dispatch_command(output_toy_data)



import autograd.numpy as np

from autograd import grad
from autograd.util import flatten
from scipy.optimize import fmin_l_bfgs_b

from diagonal_gaussian_multivariate import (
    init_natural_param_dict_from_data,
    calc_log_proba_data_arr_NK,
    mu2eta,
    eta2mu,
)

def make_problem(n_examples=1000, seed=0):
    prng = np.random.RandomState(seed)
    n_dims = 6
    mu_DK = np.asarray([[-2, -1, 0.0, 1, 2, 5]]).T
    stddev_DK = np.asarray([[0.1, 1.0, 10.0, 1.0, 0.1, 0.001]]).T
    data_ND = mu_DK[:, 0] + stddev_DK[:, 0] * prng.randn(n_examples, n_dims)
    data_DN = data_ND.T
    return data_DN, dict(
        mu_DK=mu_DK, stddev_DK=stddev_DK)

if __name__ == '__main__':
    np.set_printoptions(precision=3, suppress=1)

    data_DN, true_param_dict = make_problem(n_examples=10000, seed=0)

    init_param_dict = init_natural_param_dict_from_data(data_DN)
    randinit_flat_vec, unflatten_func = flatten(init_param_dict)
    trueinit_flat_vec, _ = flatten(mu2eta(**true_param_dict))

    def calc_neg_log_proba_dataset_wrt_flat_eta(flat_vector):
        param_dict = unflatten_func(flat_vector)
        return -1 * np.sum(
            calc_log_proba_data_arr_NK(data_DN, **param_dict))
    grad_of_neg_log_proba_dataset_wrt_flat_eta = \
        grad(calc_neg_log_proba_dataset_wrt_flat_eta)

    fromrand_flat_vec, fromrand_loss_val, fromrand_dict = fmin_l_bfgs_b(
        calc_neg_log_proba_dataset_wrt_flat_eta,
        randinit_flat_vec,
        fprime=grad_of_neg_log_proba_dataset_wrt_flat_eta)
    fromrand_param_dict = eta2mu(**unflatten_func(fromrand_flat_vec))

    fromtrue_flat_vec, fromtrue_loss_val, fromtrue_dict = fmin_l_bfgs_b(
        calc_neg_log_proba_dataset_wrt_flat_eta,
        trueinit_flat_vec,
        fprime=grad_of_neg_log_proba_dataset_wrt_flat_eta)
    fromtrue_param_dict = eta2mu(**unflatten_func(fromtrue_flat_vec))

    def pprint_param_dict(label='', mu_DK=None, stddev_DK=None):
        print label
        print  'mean vec   ' + \
            ' '.join(['% 9.5f' % a for a in mu_DK.flatten()])
        print 'stddev vec ' + \
            ' '.join(['% 9.5f' % a for a in stddev_DK.flatten()])

    pprint_param_dict('TRUE PARAMS', **true_param_dict)
    pprint_param_dict('FROMRAND ESTIMATED PARAMS', **fromrand_param_dict)
    print fromrand_loss_val

    pprint_param_dict('FROMTRUE ESTIMATED PARAMS', **fromtrue_param_dict)
    print fromtrue_loss_val


'''
Bernoulli Multivariate observation model

Parameters
----------
K : num states
D : num dimensions for each observation
eta_KD : NATURAL parameters
    eta_KD[k, d] = log odds ratio of observing dim d with cluster k
'''

import autograd.numpy as np
import sys
import os
import pdb

from autograd.scipy.misc import logsumexp

def init_natural_param_dict_from_data(
        data_DN, K=1, n_states=None, seed=0):
    '''
    '''
    prng = np.random.RandomState(int(seed))
    if n_states is not None:
        K = int(n_states)
    else:
        K = int(K)
    D = data_DN.shape[0]
    eta_DK = prng.randn(K, D).T
    return dict(
        eta_DK=eta_DK)


def init_natural_param_array(D, K, seed=0):
    prng = np.random.RandomState(int(seed))
    D, K = int(D), int(K)
    eta_DK = prng.randn(K, D).T
    return eta_DK


def eta2mu(eta_DK):
    ''' Convert mean to natural parameters

    Returns
    -------
    mu_DK : 2D array, D x K

    Examples
    --------
    >>> np.set_printoptions(suppress=0, precision=3)
    >>> eta = np.asarray([-4, -1, 0, 1, 4])
    >>> print eta2mu(eta)
    [ 0.018  0.269  0.5    0.731  0.982]
    '''
    return 1.0 / (1.0 + np.exp(-eta_DK))

def mu2eta(mu_DK):
    ''' Convert mean to natural parameters

    Returns
    -------
    eta_DK : 2D array, D x K

    Examples
    --------
    >>> np.set_printoptions(suppress=0, precision=3)
    >>> mu = np.asarray([0.018, 0.269, 0.5, 0.731, 0.982])
    >>> print mu2eta(mu)
    [-3.999 -1.     0.     1.     3.999]
    '''
    mu_DK = np.maximum(np.minimum(mu_DK, 0.99999), 0.00001)
    return np.log( mu_DK / (1 - mu_DK) )


def calc_log_proba_data_arr_NK(data_DN, eta_DK=None, mu_DK=None, **kwargs):
    ''' Compute log probability of each dataset observation under each cluster

    Args
    ----
    data_DN : 2D array, n_dims x n_examples, with binary values
        data_ND[n, d] = 1 if dimension d is ON for observation n
    eta_DK : 2D array, n_dims x n_states , with real values
        log odds ratio for observing ON value under cluster k and dim d
        not needed if mu_KD provided
    mu_DK : 2D array, n_dims x n_states, with values 0 < mu < 1
        probability of observing ON value under cluster k and dim d

    Returns
    -------
    log_proba_NK : 2D array, n_examples x n_states, with real values
        entry n,k gives log probability of data row n under state k

    Examples
    --------
    >>> np.set_printoptions(suppress=0, precision=3)
    >>> data_DN = np.asarray([[0, 1]])
    >>> mu_DK = np.asarray([[0.8, 0.1]])
    >>> log_proba_NK_v1 = calc_log_proba_data_arr_NK(
    ...     data_DN, mu_DK=mu_DK)
    >>> log_proba_NK_v2 = calc_log_proba_data_arr_NK(
    ...     data_DN, eta_DK=mu2eta(mu_DK))

    # Verify that the log probas are < 0
    >>> assert np.all(log_proba_NK_v1 < 0)

    # Verify that the log probas sum to 1 across 0 and 1
    >>> assert np.allclose(1.0, np.sum(np.exp(log_proba_NK_v1), axis=0))

    # Verify two methods are the same
    >>> assert np.allclose(log_proba_NK_v1, log_proba_NK_v2)

    >>> prng = np.random.RandomState(0)
    >>> big_data_DN = prng.rand(3, 100) > 0.3
    >>> mu_DK = prng.rand(3, 2)
    >>> log_proba_NK_v1 = calc_log_proba_data_arr_NK(
    ...     big_data_DN, mu_DK=mu_DK)
    >>> log_proba_NK_v2 = calc_log_proba_data_arr_NK(
    ...     big_data_DN, eta_DK=mu2eta(mu_DK))

    # Verify two methods are the same
    >>> assert np.allclose(log_proba_NK_v1, log_proba_NK_v2)
    '''
    if mu_DK is not None:
        # Using mean parameter
        log_mu_DK = np.log(mu_DK)
        log_1minus_mu_DK = np.log(1 - mu_DK)
        log_proba_NK = np.dot(data_DN.T, log_mu_DK)
        log_proba_NK += np.dot(1 - data_DN.T, log_1minus_mu_DK)
    else:
        # Using natural parameter
        log_cumulant_K = np.sum(np.log(1 - eta2mu(eta_DK)), axis=0)
        log_proba_NK = np.dot(data_DN.T, eta_DK)
        log_proba_NK += log_cumulant_K[np.newaxis, :]
    return log_proba_NK



"""
Similar file to early_stopping/optimize.py but
depends on converging vs early_stopping
"""

from __future__ import absolute_import

import autograd.numpy as np
from autograd.util import flatten_func
from builtins import range
from random import sample
from math import ceil
from copy import copy
import pdb


def adam(
        grad,
        init_params,
        num_batches=1,
        callback=None,
        min_epoch=1,
        max_epoch=20,
        step_size=0.001,
        b1=0.9,
        b2=0.999,
        eps=10**-8,
        err_fun=None,
        va_ll_fun=None,
        va_err_fun=None,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10):
    """ Adam as described in http://arxiv.org/pdf/1412.6980.pdf.
    It's basically RMSprop with momentum and some correction terms."""

    flattened_grad, unflatten, x = flatten_func(grad, init_params)

    # initial settings for variables
    m, v = np.zeros(len(x)), np.zeros(len(x))
    reset_patience = patience

    # initialize loglikelihoods (these are used to determine convergence)
    # we define a single ll per batch so we know which to compare to.
    # Comparing ll between batches doesn't make sense.
    old_ll, cur_ll = np.ones(num_batches), np.zeros(num_batches)
    cur_va_ll, best_va_ll = None, None

    cur_err = np.zeros(num_batches)
    cur_va_err = None

    # training goes until all batches have converged / max_iter
    have_converged = np.zeros(num_batches)

    cur_iter = 0  # != epoch
    cur_batch = 0
    cur_epoch = 0

    while (cur_epoch < max_epoch) or (cur_epoch < min_epoch):
        # we can test convergence for every batch and keep track
        # of which batches have converged
        if (np.abs(cur_ll[cur_batch] - old_ll[cur_batch]) <= stop_criterion):
            have_converged[cur_batch] = 1
        else:  # this should rarely fire
            have_converged[cur_batch] = 0

        # if all batches have converged before max_iter: break
        if sum(have_converged) == num_batches:
            break

        fg = flattened_grad(x, cur_iter)  # pass iter for batch training
        ll, g = fg[0], fg[1:]

        m = (1 - b1) * g + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g**2) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1**(cur_iter + 1))    # Bias correction.
        vhat = v / (1 - b2**(cur_iter + 1))
        x = x - step_size*mhat/(np.sqrt(vhat) + eps)

        old_ll[cur_batch] = cur_ll[cur_batch]
        cur_ll[cur_batch] = ll

        if err_fun:
            cur_err[cur_batch] = err_fun(x, cur_iter)

        if va_ll_fun and cur_batch == num_batches - 1:
            cur_va_ll = va_ll_fun(x)

            # stop based on va ll
            if early_stop:
                if cur_epoch == 0:
                    best_va_ll = cur_va_ll
                    best_x = x
                else:
                    if cur_va_ll <= best_va_ll:
                        best_va_ll = cur_va_ll
                        best_x = x
                        patience = reset_patience
                    else:
                        patience -= 1

                if patience <= 0:
                    break

            if va_err_fun:
                cur_va_err = va_err_fun(x)

        if callback:
            callback(cur_epoch,
                     cur_batch,
                     np.sum(cur_ll[:cur_batch+1]),
                     np.mean(cur_err[:cur_batch+1]),
                     cur_va_ll,
                     cur_va_err,
                     x)

        cur_iter += 1
        cur_batch = cur_iter % num_batches
        cur_epoch = cur_iter // num_batches

    return unflatten(best_x if early_stop else x)



import numpy as np
from rlstm.nn_util import sigmoid, add_bias_and_multiply
import time

def benchmark_runtime_for_func(func, *args, **kwargs):
    if 'n_reps' in kwargs:
        n_reps = kwargs['n_reps']
        del kwargs['n_reps']
    else:
        n_reps = 1
    elapsed_time_sec_per_rep = np.zeros(n_reps)
    for rep in xrange(n_reps):
        start_time_sec = time.time()
        func(*args, **kwargs)
        elapsed_time_sec_per_rep[rep] = time.time() - start_time_sec

    print "n_reps %d  median sec %9.4f  max sec %9.4f  %s" % (
        n_reps,
        np.median(elapsed_time_sec_per_rep),
        np.max(elapsed_time_sec_per_rep),
        func.__name__,
        )
    return np.mean(elapsed_time_sec_per_rep), elapsed_time_sec_per_rep

def simple_update_for_single_timestep_with_preweighted_input(
        curr_weighted_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights):
    update = sigmoid(
        curr_weighted_input +
        np.dot(prev_hiddens, update_h_weights))
    hiddens = (1 - update) * prev_hiddens
    return hiddens

def simple_update_for_single_timestep(
        curr_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights):
    update = sigmoid(
        add_bias_and_multiply(update_x_weights, curr_input) +
        np.dot(prev_hiddens, update_h_weights))
    hiddens = (1 - update) * prev_hiddens
    return hiddens

def calc_loss_one_sequence_some_vectorization(
        curr_input=None,
        hiddens=None,
        update_x_weights=None,
        update_h_weights=None):
    weighted_input = add_bias_and_multiply(update_x_weights, curr_input.T)
    loss = 0.0
    time_count = curr_input.shape[1]
    for time_iter in range(time_count):
        hiddens = simple_update_for_single_timestep_with_preweighted_input(
            weighted_input[time_iter],
            hiddens,
            update_x_weights,
            update_h_weights)
        loss = loss + np.sum(hiddens)
    return loss

def calc_loss_one_sequence_forloop(
        curr_input=None,
        hiddens=None,
        update_x_weights=None,
        update_h_weights=None):
    loss = 0.0
    time_count = curr_input.shape[1]
    for time_iter in range(time_count):
        hiddens = simple_update_for_single_timestep(
            np.expand_dims(curr_input[:, time_iter], axis=0),
            hiddens,
            update_x_weights,
            update_h_weights)
        loss = loss + np.sum(hiddens)
    return loss

if __name__ == '__main__':
    T = 20000
    D = 6
    H = 100
    prng = np.random.RandomState(0)
    curr_input = prng.randn(D, T)
    update_x_weights = prng.randn(D+1, H)
    update_h_weights = prng.randn(H, H)
    hiddens = prng.randn(1, H)

    loss1 = calc_loss_one_sequence_forloop(
        curr_input,
        hiddens,
        update_x_weights,
        update_h_weights)
    print loss1

    loss2 = calc_loss_one_sequence_some_vectorization(
        curr_input,
        hiddens,
        update_x_weights,
        update_h_weights)
    print loss2

    args = (curr_input, hiddens, update_x_weights, update_h_weights)

    # Time the old-school forloop over each timestep
    benchmark_runtime_for_func(
        calc_loss_one_sequence_forloop,
        *args,
        n_reps=3)

    # Time the newer version, which does some vectorization
    benchmark_runtime_for_func(
        calc_loss_one_sequence_some_vectorization,
        *args,
        n_reps=3)



from __future__ import absolute_import, print_function

import pdb
from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.weights_parser import WeightsParser
from rlstm.nn_util import sigmoid, add_bias_and_multiply
from rlstm.common_util import safe_log

# from debug_util import Timer, getCurMemCost_MiB

# -----------------------
#    GRU DROPOUT UTILS
# -----------------------

def kdropout(weights, k):
    ''' chooses k indexes of x. set others to 0. '''
    num_inputs = weights.shape[0]
    num_states = weights.shape[1]
    new_weights = copy(weights)
    # sort the weights
    sort_id_weights = np.argsort(np.abs(new_weights), axis=1)
    sort_weights = new_weights[np.arange(num_inputs)[:, None], sort_id_weights]
    sort_weights = np.hstack((np.zeros((num_inputs, num_states-k)), sort_weights[:, -k:]))
    # recover indexes into new_weights
    recover_sort_id = np.argsort(sort_id_weights, axis=1)
    new_weights = sort_weights[np.arange(num_inputs)[:, None], recover_sort_id]
    return new_weights

# ----------------------
#    GRU REDUN UTILS
# ----------------------

def row_norms(X):
    norms = np.einsum('ij,ij->i', X, X)
    return np.sqrt(norms)


def cosine_cdist(XA, XB):
    mA = XA.shape[0]
    mB = XB.shape[0]
    dm = np.zeros((mA, mB), dtype=np.double)

    normsA = row_norms(XA)
    normsB = row_norms(XB)

    dm = np.dot(XA, XB.T)
    dm /= normsA.reshape(-1, 1)
    dm /= normsB
    dm *= -1
    dm += 1
    return dm


def sparse_cos_dists(A, B):
    cos_dists_output = cosine_cdist(A, B)
    # symmetric, so just keep 1/2 (ignore diagonal)
    return np.triu(cos_dists_output, k=1)


def is_redundant(A, epsilon=0.5):
    redun_idx = []
    idx = np.argwhere(np.logical_and(np.abs(A) < epsilon, A != 0))
    if idx.shape[0] > 0:  # get unique so no duplicate removals
        # quick way to get rid of duplicates
        idx = idx[np.unique(idx[:, 1], return_index=True)[1], :]
        # create a map from uid1 to uid2
        uid_pool = np.unique(idx[:, 0])
        for uid in uid_pool:
            redun_idx.append((uid, idx[:, 1][idx[:, 0] == uid]))
    return redun_idx, idx[:, 1]


def autograd_zero(A, index):
    ''' Because autograd can't support indexing, we have to
        be kind of clever about it
            - this handles empty indexes
    '''
    B = copy(A.T)
    if index.shape[0] > 0:
        index_all = np.arange(B.shape[1])
        index_ordered = np.concatenate(
            (index, index_all[~np.in1d(index_all, index)]))
        B_ordered = B[np.arange(B.shape[0])[:, None], index_ordered]
        B_zeroed = np.concatenate(
            (np.zeros((B.shape[0], len(index))), B_ordered[:, len(index):]), axis=1)
        index_recover = np.argsort(index_ordered)
        B = B_zeroed[np.arange(B.shape[0])[:, None], index_recover]
    return B.T


def autograd_add(A, index):
    B = copy(A)
    if len(index) > 0:
        for i, j in index:
            C = np.expand_dims(B[i, :] + np.sum(B[j, :], axis=0), axis=0)
            B = np.concatenate((B[:i, :], C, B[i+1:, :]))
    return B


def split_weights(weights, num_hiddens):
    num_parts = weights.shape[1] / num_hiddens
    return [weights[:, i*num_hiddens:(i+1)*num_hiddens]
        for i in range(num_parts)]

# --------------------------
#    GRU UPDATE FUNCTIONS
# --------------------------

def update(
        max_conn,
        redun_epsilon,
        curr_weighted_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights,
        reset_x_weights,
        reset_h_weights,
        thidden_x_weights,
        thidden_h_weights
    ):

    if redun_epsilon > 0:
        num_hiddens = prev_hiddens.shape[1]

        # get "redundant" index by cosine distances between weight matrices
        all_x_weights = np.concatenate(
            (update_x_weights, reset_x_weights, thidden_x_weights), axis=1)
        all_x_redun, xy_redun = is_redundant(
            sparse_cos_dists(all_x_weights, all_x_weights), redun_epsilon)

        all_h_weights = np.concatenate(
            (update_h_weights, reset_h_weights, thidden_h_weights), axis=1)
        all_h_redun, hy_redun = is_redundant(
            sparse_cos_dists(all_h_weights, all_h_weights), redun_epsilon)

        # 0 out the proper indexes in the connections
        all_x_weights = autograd_zero(
            autograd_add(all_x_weights, all_x_redun), xy_redun)
        all_h_weights = autograd_zero(
            autograd_add(all_h_weights, all_h_redun), hy_redun)

        # get the individual vectors out
        update_x_weights, reset_x_weights, thidden_x_weights = \
            split_weights(all_x_weights, num_hiddens)
        update_h_weights, reset_h_weights, thidden_h_weights = \
            split_weights(all_h_weights, num_hiddens)

    if max_conn > 0:
        update_x_weights = kdropout(update_x_weights, max_conn)
        update_h_weights = kdropout(update_h_weights, max_conn)
        reset_x_weights = kdropout(reset_x_weights, max_conn)
        reset_h_weights = kdropout(reset_h_weights, max_conn)
        thidden_x_weights = kdropout(thidden_x_weights, max_conn)
        thidden_h_weights = kdropout(thidden_h_weights, max_conn)

    update = sigmoid(curr_weighted_input + np.dot(prev_hiddens, update_h_weights))
    reset = sigmoid(curr_weighted_input + np.dot(prev_hiddens, reset_h_weights))
    thiddens = np.tanh(curr_weighted_input + np.dot(reset * prev_hiddens, thidden_h_weights))
    hiddens = (1 - update) * prev_hiddens + update * thiddens

    return hiddens

# -------------------
#   GRU CONSTRUCTOR
# -------------------

def build(
        input_count,
        state_count,
        output_count,
        max_conn=0,
        redun_epsilon=0
    ):
    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)

    # max_conn specifies max-dropout probabilities
    # redun_epsilon specifies unit merging probabilities

    parser = WeightsParser()
    parser.add_shape('init_hiddens', (1, state_count))
    parser.add_shape('update_x_weights', (input_count + 1, state_count))
    parser.add_shape('update_h_weights', (state_count, state_count))
    parser.add_shape('reset_x_weights', (input_count + 1, state_count))
    parser.add_shape('reset_h_weights', (state_count, state_count))
    parser.add_shape('thidden_x_weights', (input_count + 1, state_count))
    parser.add_shape('thidden_h_weights', (state_count, state_count))
    parser.add_shape('output_h_weights', (state_count, output_count))

    def outputs(
            weights,
            input_set,
            fence_set,
            output_set=None,
            return_pred_set=False
        ):

        update_x_weights = parser.get(weights, 'update_x_weights')
        update_h_weights = parser.get(weights, 'update_h_weights')
        reset_x_weights = parser.get(weights, 'reset_x_weights')
        reset_h_weights = parser.get(weights, 'reset_h_weights')
        thidden_x_weights = parser.get(weights, 'thidden_x_weights')
        thidden_h_weights = parser.get(weights, 'thidden_h_weights')
        output_h_weights = parser.get(weights, 'output_h_weights')

        data_count = len(fence_set) - 1
        feat_count = input_set.shape[0]

        ll = 0.0
        n_i_track = 0
        fence_base = fence_set[0]
        pred_set = None

        if return_pred_set:
            pred_set = np.zeros((output_count, input_set.shape[1]))

        # loop through sequences and time steps
        for data_iter in range(data_count):
            hiddens = copy(parser.get(weights, 'init_hiddens'))

            fence_post_1 = fence_set[data_iter] - fence_base
            fence_post_2 = fence_set[data_iter+1] - fence_base
            time_count = fence_post_2 - fence_post_1
            curr_input = input_set[:, fence_post_1:fence_post_2]
            # add bias and multiply weights into inputs all together
            weighted_input = add_bias_and_multiply(update_x_weights, curr_input.T).T

            for time_iter in range(time_count):
                # print("[seq {}, timestep {}, memcost {} MB]".format(data_iter, time_iter, getCurMemCost_MiB()))
                hiddens = update(
                    max_conn,
                    redun_epsilon,
                    np.expand_dims(weighted_input[:, time_iter], axis=0),
                    hiddens,
                    update_x_weights,
                    update_h_weights,
                    reset_x_weights,
                    reset_h_weights,
                    thidden_x_weights,
                    thidden_h_weights
                )

                if output_set is not None:
                    # subtract a small number so -1
                    out_proba = sigmoid(np.sign(output_set[:, n_i_track] - 1e-3) * np.dot(hiddens, output_h_weights))
                    out_lproba = safe_log(out_proba)
                    ll += np.sum(out_lproba)
                else:
                    out_proba = sigmoid(np.dot(hiddens, output_h_weights))
                    out_lproba = safe_log(out_proba)

                if return_pred_set:
                    pred_set[:, n_i_track] = out_lproba[0]

                n_i_track += 1

        return ll, pred_set

    def prediction(weights, input_set, fence_set):
        _, output_set = outputs(weights, input_set, fence_set, return_pred_set=True)
        return output_set

    def log_likelihood(weights, input_set, fence_set, output_set):
        ll, _ = outputs(
            weights,
            input_set,
            fence_set,
            output_set=output_set,
            return_pred_set=False
        )
        return ll

    return prediction, log_likelihood, parser.num_weights



from __future__ import print_function
from collections import defaultdict

import autograd.numpy as np
from autograd import value_and_grad, grad

from rlstm.optimize import adam
from rlstm import create_toy_data

from rlstm.models import lr, hmm, mlp, gru, gruhmm
from rlstm.interpret.rnn_tree import train_decision_tree
from rlstm.debug_util import getCurMemCost_MiB

from rlstm.nn_util import (init_hmm_emission_weights,
                           init_hmm_out_weights,
                           init_vanilla_gru_weights)

from rlstm.common_util import (safe_log,
                               flatten_to_2d,
                               wrapper_func,
                               map_3d_to_2d)

from rlstm.scores import get_accuracy, get_mean_squared_error
from rlstm.autograd_util import autograd_extract, is_autograd_array

from copy import copy
import cPickle
import dill
import argparse
from datetime import datetime

import pdb  # remove me

# -- train functions --


def train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        logpred_fun,
        num_weights,
        out_set=None,
        init_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.01,
        # regularization
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        num_l1_weights=0,
        num_l2_weights=0,
        # optimization
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        regression=False,
        # validation set
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None,
        # overwrite functions
        batch_indices=None,
        regularized_loss=None,
        batch_loss=None,
        batch_error=None,
        callback=None):
    ''' Generic function for training using stochastic gradient descent.
        All models are wrapped around this function.
    '''

    # assume the *num_data* is always last index
    num_data = len(fence_set) - 1
    num_batches = int(np.ceil(num_data / float(batch_size)))

    if init_weights is None:
        init_weights = np.random.randn(num_weights) * param_scale

    # choose the proper error function
    error_func = get_mean_squared_error if regression else get_accuracy

    if batch_indices is None:
        def batch_indices(iter):  # iter = batch_iter
            idx = iter % num_batches
            return slice(idx * batch_size, (idx+1) * batch_size)

    # wrapper around loss function
    if regularized_loss is None:
        def regularized_loss(weights, obs_set, fence_set, out_set):
            loss = loglike_fun(weights, obs_set, fence_set, out_set)
            if out_set is not None and obs_lambda > 0:
                # Add a parameter to balance between out log-likelihood and obs log-likelihood
                loss += obs_lambda*loglike_fun(weights, obs_set, fence_set, None)

            # negative log like + regularizers
            loss = -loss + \
                l1_lambda * np.sum(np.abs(weights[:num_l1_weights])) + \
                l2_lambda * np.sum(np.power(weights[:num_l2_weights], 2))
            return loss

    def _get_batch_sets(idx):
         # find the relevant fence posts
        batch_i = np.arange(num_data)[idx]
        batch_fence_set = np.unique(np.concatenate((fence_set[batch_i], fence_set[batch_i+1])))

        # use the fence sets to get batch obs/out set
        batch_out_set = None
        for i in range(batch_i.size):
            if i == 0:
                batch_obs_set = obs_set[:, batch_fence_set[i]:batch_fence_set[i+1]]
                if not out_set is None:
                    batch_out_set = out_set[:, batch_fence_set[i]:batch_fence_set[i+1]]
            else:
                cur_batch_obs_set = obs_set[:, batch_fence_set[i]:batch_fence_set[i+1]]
                batch_obs_set = np.concatenate((batch_obs_set, cur_batch_obs_set), axis=1)
                if not out_set is None:
                    cur_batch_out_set = out_set[:, batch_fence_set[i]:batch_fence_set[i+1]]
                    batch_out_set = np.concatenate((batch_out_set, cur_batch_out_set), axis=1)

        return batch_obs_set, batch_fence_set, batch_out_set

    # loss function in batches
    if batch_loss is None:
        def batch_loss(weights, iter):
            idx = batch_indices(iter)
            batch_obs_set, batch_fence_set, batch_out_set = _get_batch_sets(idx)
            return regularized_loss(weights, batch_obs_set, batch_fence_set, batch_out_set)

    if batch_error is None:
        def batch_error(weights, iter):
            idx = batch_indices(iter)
            batch_obs_set, batch_fence_set, batch_out_set = _get_batch_sets(idx)
            batch_logpred_set = logpred_fun(weights, batch_obs_set, batch_fence_set)
            batch_pred_set = np.exp(batch_logpred_set)

            if batch_out_set is not None:
                return error_func(batch_out_set, batch_pred_set)
            return None

    if callback is None:
        def callback(e_i, b_i, ll, acc, va_ll, va_acc, x):
            print('epoch {:_<4}  |  batch {:_<2}/{:_<2}  |  loss {:_<10}  |  acc {:_<6}'.format(
                e_i,
                b_i+1,
                num_batches,
                round(ll, 4),
                round(acc, 4)))
            if b_i == num_batches - 1:
                print('======================================================================================================')
                print('EPOCH SUMMARY - LOSS {:_<10} | ACC {:_<6} | VA_LOSS {:_<10} | VA_ACC {:_<6} | MEM {}'.format(
                    round(ll, 4),
                    round(acc, 4),
                    round(va_ll, 4) if va_ll else 'nan',
                    round(va_acc, 4) if va_acc else 'nan',
                    int(getCurMemCost_MiB())))
                print('======================================================================================================')

    grad_fun = value_and_grad(batch_loss)
    weights = adam(grad_fun,
                   init_weights,
                   step_size=0.001,
                   num_batches=num_batches,
                   min_epoch=min_epoch,
                   max_epoch=max_epoch,
                   callback=callback,
                   stop_criterion=stop_criterion,
                   err_fun=lambda x, i: batch_error(x, i),
                   va_ll_fun=lambda x: regularized_loss(x, va_obs_set, va_fence_set, va_out_set) if va_out_set is not None else None,
                   va_err_fun=lambda x: error_func(va_out_set, np.exp(logpred_fun(x, va_obs_set, va_fence_set))) if va_out_set is not None else None,
                   early_stop=early_stop,
                   patience=patience)

    return weights


def train_lr(
        obs_set,
        fence_set,
        out_set,
        init_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.0001,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None):

    in_dim_count = obs_set.shape[0]
    out_dim_count = out_set.shape[0]

    num_weights = np.prod((out_dim_count, in_dim_count + 1))
    loglike_fun = lr.log_likelihood
    logpred_fun = lambda weights, obs_set, fence_set: lr.logistic_log_predictions(
        weights, obs_set, fence_set, in_dim_count, out_dim_count)
    # pred_fun = wrapper_func(logpred_fun, np.exp)

    trained_weights = train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        logpred_fun,
        num_weights,
        out_set=out_set,
        init_weights=init_weights,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        obs_lambda=obs_lambda,
        l1_lambda=l1_lambda,
        num_l1_weights=num_weights,
        num_l2_weights=num_weights,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_fence_set=va_fence_set,
        va_out_set=va_out_set)

    return logpred_fun, loglike_fun, trained_weights


def append_lr_to_model(
        obs_set,
        fence_set,
        out_set,
        hmm_weights,
        output_fun,
        loglike_fun,
        init_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.0001,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None):
    ''' Often we train a generative model for an hmm but want to
        tag an lr afterwards. This is wrapper function around train_lr
        to do so.
    '''

    num_hmm_weights = hmm_weights.shape[0]
    lprob_set = hmm.get_state_prob_set(
        output_fun, hmm_weights, obs_set, fence_set)

    va_lprob_set = None
    if not va_obs_set is None:
        va_lprob_set = hmm.get_state_prob_set(
            output_fun, hmm_weights, va_obs_set, va_fence_set)

    # train lr
    logpred_fun, ll_fun, lr_weights = train_lr(lprob_set,
                                               fence_set,
                                               out_set,
                                               init_weights=init_weights,
                                               min_epoch=min_epoch,
                                               max_epoch=max_epoch,
                                               batch_size=batch_size,
                                               param_scale=param_scale,
                                               obs_lambda=obs_lambda,
                                               l1_lambda=l1_lambda,
                                               stop_criterion=stop_criterion,
                                               early_stop=early_stop,
                                               patience=patience,
                                               va_obs_set=va_lprob_set,
                                               va_fence_set=va_fence_set,
                                               va_out_set=va_out_set)

    trained_weights = np.hstack((hmm_weights, lr_weights))

    # return functions so users can pass in obs_farray (instead of prob_farray)
    def all_logpred_fun(weights, obs_set, fence_set):
        lprob_set = hmm.get_state_prob_set(
            output_fun, weights, obs_set, fence_set)
        hmm_weights = weights[:num_hmm_weights]
        lr_weights = weights[num_hmm_weights:]
        return logpred_fun(lr_weights, lprob_set, fence_set)

    def all_loglike_fun(weights, obs_set, fence_set, out_set=None):
        ''' if out_farray, use lr proba; else use hmm proba
        '''

        # we know but exist (but just may want to use one)
        hmm_weights = weights[:num_hmm_weights]
        lr_weights = weights[num_hmm_weights:]

        if not out_set is None:
            lprob_set = hmm.get_state_prob_set(
                output_fun, hmm_weights, obs_set, fence_set)
            return ll_fun(lr_weights, lprob_set, fence_set, out_set)
        else:
            return loglike_fun(hmm_weights, obs_set, fence_set)

    return all_logpred_fun, all_loglike_fun, trained_weights


def train_hmm_sgd(
        obs_set,
        fence_set,
        out_set,
        hmm_state_count=5,
        emission_type='bernoulli',
        obs_only=False,
        init_hmm_weights=None,
        init_lr_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.01,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None):
    ''' Train an HMM with stochastic gradient descent:
        If obs_only, then we just optimize the likelihood
        of the observations in obs_farray. Otherwise, the out_farray
        is provided, and the hmm uses both in its cost function.
    '''

    obs_count = obs_set.shape[0]
    out_count = None
    if not obs_only:
        out_count = out_set.shape[0]

    # Build the model
    outputs, loglike_fun, num_weights = hmm.build(
        obs_count, hmm_state_count, out_count, emission_type=emission_type)

    # Initializing the weights randomly, but making sure
    # that they still represent distributions, etc.
    if init_hmm_weights is None:
        init_hmm_weights = init_hmm_emission_weights(
            hmm_state_count, obs_count, flatten=True,
                                        emission_type=emission_type)

    if init_lr_weights is None:
        init_lr_weights = init_hmm_out_weights(
            hmm_state_count, out_set.shape[0], param_scale)

    if not obs_only:
        init_weights = np.hstack((init_hmm_weights, init_lr_weights))
    else:
        init_weights = init_hmm_weights

    trained_weights = train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        outputs,
        num_weights,
        out_set=None if obs_only else out_set,
        init_weights=init_weights,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        obs_lambda=obs_lambda,
        l1_lambda=l1_lambda,
        num_l1_weights=num_weights,
        num_l2_weights=num_weights,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_fence_set=va_fence_set,
        va_out_set=None if obs_only else va_out_set)

    if obs_only:
        logpred_fun, loglike_fun, trained_weights = \
            append_lr_to_model(obs_set,
                               fence_set,
                               out_set,
                               trained_weights,
                               outputs,
                               loglike_fun,
                               init_weights=init_lr_weights,
                               min_epoch=min_epoch,
                               max_epoch=max_epoch,
                               batch_size=batch_size,
                               param_scale=param_scale,
                               obs_lambda=obs_lambda,
                               l1_lambda=l1_lambda,
                               stop_criterion=stop_criterion,
                               early_stop=early_stop,
                               patience=patience,
                               va_obs_set=va_obs_set,
                               va_fence_set=va_fence_set,
                               va_out_set=va_out_set)

        pred_fun = wrapper_func(logpred_fun, np.exp)
    else:
        pred_fun = wrapper_func(outputs, np.exp)

    return pred_fun, loglike_fun, trained_weights


def train_mlp(
        obs_set,
        out_set,
        states_list,
        init_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.01,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_out_set=None):

    from _models_util import _train_by_sgd
    input_count = obs_set.shape[0]
    output_count = out_set.shape[0]

    logpred_fun, loglike_fun, num_weights = mlp.build(
        input_count, states_list, output_count)
    pred_fun = wrapper_func(logpred_fun, np.exp)

    trained_weights = _train_by_sgd(
        obs_set,
        loglike_fun,
        logpred_fun,
        num_weights,
        out_set=out_set,
        init_weights=init_weights,
        regression=True,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_out_set=va_out_set)

    return pred_fun, loglike_fun, trained_weights


def train_gru(
        obs_set,
        fence_set,
        out_set,
        gru_state_count,
        init_weights=None,
        min_epoch=1,
        max_epoch=20,
        batch_size=128,
        param_scale=0.01,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        max_conn=0,
        redun_epsilon=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None):

    input_count = obs_set.shape[0]
    output_count = out_set.shape[0]

    # Build the model
    logpred_fun, loglike_fun, num_weights = gru.build(
        input_count, gru_state_count, output_count,
        max_conn=max_conn, redun_epsilon=redun_epsilon)

    pred_fun = wrapper_func(logpred_fun, np.exp)

    trained_weights = train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        logpred_fun,
        num_weights,
        out_set=out_set,
        init_weights=init_weights,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        obs_lambda=obs_lambda,
        l1_lambda=l1_lambda,
        num_l1_weights=num_weights,
        num_l2_weights=num_weights,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_fence_set=va_fence_set,
        va_out_set=va_out_set)

    return pred_fun, loglike_fun, trained_weights


def train_gru_hmm(
        obs_set,
        fence_set,
        out_set,
        hmm_state_count,
        gru_state_count,
        emission_type='bernoulli',
        init_gru_weights=None,
        init_hmm_weights=None,
        init_weights=None,
        min_epoch=1,
        max_epoch=100,
        batch_size=128,
        param_scale=0.01,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        max_conn=0,
        redun_epsilon=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None):

    input_count = obs_set.shape[0]
    output_count = out_set.shape[0]

    # Build the model
    logpred_fun, loglike_fun, num_gru_weights, num_hmm_weights = gruhmm.build(
        input_count,
        hmm_state_count,
        gru_state_count,
        output_count,
        max_conn=max_conn,
        redun_epsilon=redun_epsilon)

    pred_fun = wrapper_func(logpred_fun, np.exp)

    if init_gru_weights is None:
        init_gru_weights = init_vanilla_gru_weights(gru_state_count,
                                                    input_count,
                                                    output_count,
                                                    param_scale)

    if init_hmm_weights is None:
        init_hmm_weights = np.hstack((
            init_hmm_emission_weights(
                hmm_state_count,
                input_count,
                flatten=True,
                emission_type=emission_type),
            init_hmm_out_weights(
                hmm_state_count,
                output_count,
                param_scale)))

    if init_weights is None:
        init_weights = np.hstack((init_gru_weights, init_hmm_weights))

    trained_weights = train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        logpred_fun,
        num_gru_weights + num_hmm_weights,
        out_set=out_set,
        init_weights=init_weights,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        obs_lambda=obs_lambda,
        l1_lambda=l1_lambda,
        num_l1_weights=num_gru_weights,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_fence_set=va_fence_set,
        va_out_set=va_out_set)

    return pred_fun, loglike_fun, trained_weights


# ---------------------------------------------------------------------------
# Exploratory code Tree Regularization


def _dict_get(d, key, default=None):
    return d[key] if key in d else default


class BlackBoxTreeRegularizer(object):
    ''' Add a wrapper function to regularize a black-box model using
        a WEIGHT-NODE tree mapping.

        In order for the tree to be differentiable, an MLP is trained to
        approximate the tree's WEIGHT-NODE function. The MLP is should be
        differentiated inside the model's cost function.

        This model will take return the following items:
            - tree (final iteration of the tree)
            - MLP (MLP weights)

        The black box model is assumed to have the following design:
            - prediction function: accepts (weights, inputs) and returns proba
            - log_likelihood function: accepts (weights, inputs, outputs)
                                       and returns likelihood

    '''

    def __init__(self):
        self.mlp_pred_fun = None
        self.mlp_loglike_fun = None
        self.mlp_weights = None

    def train_tree(self,
                   obs_set,
                   fence_set,
                   out_set,
                   bb_logpred_fun,
                   bb_loglike_fun,
                   bb_num_weights,
                   params={}):
        ''' Args
            ----
            obs_farray : numpy array
            out_farray : numpy array
            bb_logpred_fun : lambda function
                             rlstm.models object
            bb_loglike_fun : lambda function
                             rlstm.models object
        '''
        input_count = obs_set.shape[0]
        output_count = out_set.shape[0]
        bb_pred_fun = wrapper_func(bb_logpred_fun, np.exp)

        weights_store = np.zeros((bb_num_weights, _dict_get(params, 'max_epoch', 10000)))
        counts_store = np.zeros((1, _dict_get(params, 'max_epoch', 10000)))
        batch_size = _dict_get(params, 'batch_size', 128)

        num_data = len(fence_set) + 1
        num_batches = int(np.ceil(num_data / float(batch_size)))

        def bb_loss(weights, obs_set, fence_set, out_set):
            return bb_loglike_fun(weights, obs_set, fence_set, out_set)

        # train a tree + store stuff in the callback
        def callback(e_i, b_i, ll, acc, va_ll, va_acc, x):
            iter_i = e_i * num_batches + b_i
            tree = train_decision_tree(bb_pred_fun, x, obs_set, fence_set)
            weights_store[:, iter_i] = x
            counts_store[:, iter_i] = tree.tree_.node_count

            print('epoch {:_<4}  |  batch {:_<2}/{:_<2}  |  loss {:_<10}  |  acc {:_<5}  |  va-loss {:_<10}  |  va-acc {:_<5}'.format(
                e_i,
                b_i+1,
                num_batches,
                round(ll, 4),
                round(acc, 4),
                round(va_ll, 4) if va_ll else 'skipping',
                round(va_acc, 4) if va_acc else 'skipping'))

        trained_weights = train_by_sgd(
            obs_set,
            fence_set,
            bb_loss,
            bb_logpred_fun,
            bb_num_weights,
            out_set=out_set,
            min_epoch=_dict_get(params, 'min_epoch', 100),
            max_epoch=_dict_get(params, 'max_epoch', 10000),
            batch_size=_dict_get(params, 'batch_size', 64),
            param_scale=_dict_get(params, 'param_scale', 0.01),
            l1_lambda=_dict_get(params, 'l1_lambda', 0),
            l2_lambda=_dict_get(params, 'l2_lambda', 0),
            num_l1_weights=bb_num_weights,
            num_l2_weights=bb_num_weights,
            stop_criterion=_dict_get(params, 'stop_criterion', 1e-3),
            callback=callback
        )

        return weights_store, counts_store

    def train_mlp(self, weights, counts, num_hiddens, params={}):
        pred_fun, loglike_fun, trained_weights = train_mlp(
            weights,
            counts,
            num_hiddens,
            min_epoch=_dict_get(params, 'min_epoch', 100),
            max_epoch=_dict_get(params, 'max_epoch', 10000),
            batch_size=_dict_get(params, 'batch_size', 64),
            param_scale=_dict_get(params, 'param_scale', 0.01),
            stop_criterion=_dict_get(params, 'stop_criterion', 1e-3)
        )

        self.mlp_pred_fun = pred_fun
        self.mlp_loglike_fun = loglike_fun
        self.mlp_weights = trained_weights


def train_gru_hmm_tree(
        obs_set,
        fence_set,
        out_set,
        hmm_state_count,
        gru_state_count,
        emission_type='bernoulli',
        init_gru_weights=None,
        init_hmm_weights=None,
        init_weights=None,
        min_epoch=1,
        max_epoch=100,
        mlp_min_epoch=1,
        mlp_max_epoch=1000,
        batch_size=128,
        param_scale=0.01,
        obs_lambda=0,
        l1_lambda=0,
        l2_lambda=0,
        tree_lambda=0,
        max_conn=0,
        redun_epsilon=0,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10,
        va_obs_set=None,
        va_fence_set=None,
        va_out_set=None,
        # pretrained tree regressor
        tree_mlp_pred_fun=None,
        tree_mlp_loglike_fun=None,
        tree_mlp_weights=None,
        tree_mlp_output_loc=None):
    ''' Same as train_gru_hmm except a decision tree approximator is added
        to predict a number of nodes that acts as a regularization technique.
    '''

    input_count = obs_set.shape[0]
    output_count = out_set.shape[0]

    # Build the model
    logpred_fun, loglike_fun, num_gru_weights, num_hmm_weights = gruhmm.build(
        input_count,
        hmm_state_count,
        gru_state_count,
        output_count,
        max_conn=max_conn,
        redun_epsilon=redun_epsilon)

    pred_fun = wrapper_func(logpred_fun, np.exp)

    if init_gru_weights is None:
        init_gru_weights = init_vanilla_gru_weights(gru_state_count,
                                                    input_count,
                                                    output_count,
                                                    param_scale)

    if init_hmm_weights is None:
        init_hmm_weights = np.hstack((
            init_hmm_emission_weights(
                hmm_state_count,
                input_count,
                flatten=True,
                emission_type=emission_type),
            init_hmm_out_weights(
                hmm_state_count,
                output_count,
                param_scale)))

    if init_weights is None:
        init_weights = np.hstack((init_gru_weights, init_hmm_weights))

    # -------------------------------------------------------
    # initialize a regularizer accent
    if tree_mlp_pred_fun is None and tree_mlp_weights is None:
        treg = BlackBoxTreeRegularizer()
        _gru_logpred_fun, _gru_loglike_fun, _gru_num_weights = gru.build(
            input_count, gru_state_count, output_count)
        print('[{}] training regularization tree'.format(datetime.now()))
        tree_weights, tree_counts = treg.train_tree(
                                        obs_set,
                                        fence_set,
                                        out_set,
                                        _gru_logpred_fun,
                                        _gru_loglike_fun,
                                        _gru_num_weights,
                                        params={'min_epoch':min_epoch,
                                                'max_epoch':max_epoch,
                                                'batch_size':batch_size,
                                                'param_scale':param_scale,
                                                'stop_criterion':stop_criterion,
                                                'l1_lambda':l1_lambda,
                                                'l2_lambda':l2_lambda})

        # get rid of any excess examples
        tree_num_real = np.where(tree_counts == 0)[1]
        if tree_num_real.size > 0:
            tree_num_real = tree_num_real[0]
            tree_weights = tree_weights[:, :tree_num_real]
            tree_counts = tree_counts[:, :tree_num_real]

        print('[{}] training regularization perceptron'.format(datetime.now()))
        treg.train_mlp(tree_weights,
                       tree_counts,
                       [100],
                       params={'min_epoch':mlp_min_epoch,
                               'max_epoch':mlp_max_epoch,
                               'batch_size':batch_size,
                               'param_scale':param_scale,
                               'stop_criterion':stop_criterion})
        tree_mlp_pred_fun = treg.mlp_pred_fun
        tree_mlp_loglike_fun = treg.mlp_loglike_fun
        tree_mlp_weights = treg.mlp_weights

        if not tree_mlp_output_loc is None:
            with open(tree_mlp_output_loc, 'w') as fp:
                tree_snapshot = {'pred_fun':tree_mlp_pred_fun,
                                 'loglike_fun':tree_mlp_loglike_fun,
                                 'weights':tree_mlp_weights}
                dill.dump(tree_snapshot, fp)

    # -------------------------------------------------------

    # Overwrite to include the tree regularization.
    def regularized_loss(weights, obs_set, fence_set, out_set):
        loss = loglike_fun(weights, obs_set, fence_set, out_set)
        if not out_set is None and obs_lambda > 0:
            # Add a parameter to balance between out log-likelihood and obs log-likelihood
            loss += obs_lambda*loglike_fun(weights, obs_set, fence_set, None)

        # negative log like + regularizers
        loss = -loss + \
            l1_lambda * np.sum(np.abs(weights[:num_l1_weights])) + \
            l2_lambda * np.sum(np.power(weights[:num_l2_weights], 2))

        if tree_lambda > 0:
            # rarely MLP could predict < 0 (which makes no sense)
            # b/c we log then re-exp, < 0 --> nan
            tree_mlp_pred = tree_mlp_pred_fun(tree_mlp_weights, weights[:num_gru_weights])
            if np.isnan(tree_mlp_pred):
                tree_mlp_pred = 0
            loss += tree_lambda * np.sum(tree_mlp_pred)
        return loss

    print('[{}] training gru-hmm'.format(datetime.now()))
    trained_weights = train_by_sgd(
        obs_set,
        fence_set,
        loglike_fun,
        logpred_fun,
        num_gru_weights + num_hmm_weights,
        out_set=out_set,
        init_weights=init_weights,
        min_epoch=min_epoch,
        max_epoch=max_epoch,
        batch_size=batch_size,
        param_scale=param_scale,
        obs_lambda=obs_lambda,
        l1_lambda=l1_lambda,
        num_l1_weights=num_gru_weights,
        stop_criterion=stop_criterion,
        early_stop=early_stop,
        patience=patience,
        va_obs_set=va_obs_set,
        va_fence_set=va_fence_set,
        va_out_set=va_out_set,
        regularized_loss=regularized_loss)

    return (pred_fun, tree_mlp_pred_fun), \
           (loglike_fun, tree_mlp_loglike_fun), \
           (trained_weights, tree_mlp_weights)



import pdb
import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad, value_and_grad

from rlstm.common_util import safe_log
from rlstm.nn_util import sigmoid, flatten_io, cross_entropy


def cross_entropy(outputs, proba):
    return outputs * proba + (1 - outputs) * (1 - proba)


def logistic_predictions(weights, input_set, fence_set, in_dim_count, out_dim_count):
    # input should be ( dims by data ) or vertical columns
    # Adds in the row of ones, so we don't have to when passing things in!
    # outputs probability of a label being true according to logistic model.

    # flatten but we already have flattened version in farray
    my_input_set = np.vstack((input_set, np.ones((1, input_set.shape[1]))))
    weights = np.reshape(weights, (out_dim_count, in_dim_count + 1))
    return sigmoid(np.dot(weights, my_input_set))


def logistic_log_predictions(weights, input_set, fence_set, in_dim_count, out_dim_count):
    return safe_log(logistic_predictions(weights, input_set, fence_set, in_dim_count, out_dim_count))


def log_likelihood(weights, input_set, fence_set, output_set):
    # Training loss is the negative log-likelihood of the training labels.
    in_dim_count = input_set.shape[0]
    out_dim_count = output_set.shape[0]

    pred_set = logistic_predictions(
        weights,
        input_set,
        fence_set,
        in_dim_count,
        out_dim_count)

    lprobs = safe_log(cross_entropy(output_set, pred_set))
    ll = np.sum(lprobs)
    return ll



''' Debugging utils
        - Timing code w/o repetitive code
'''

from datetime import datetime
from collections import defaultdict
import os
import pdb
import psutil


def getCurMemCost_MiB():
    # return the memory usage in MB
    process = psutil.Process(os.getpid())
    mem_MiB = process.memory_info_ex().rss / float(2 ** 20)
    return mem_MiB


class MemTracker(object):
    def __init__(self):
        self.database = defaultdict(lambda: {
            'cost': 0,
            'avg_cost': 0,
            '_total_cost': 0,
            'n_steps': 0
        })

    def add(self, key):
        self.database[key]['cost'] = getCurMemCost_MiB()
        self.database[key]['n_steps'] += 1
        self.database[key]['_total_cost'] += self.database[key]['cost']
        self.database[key]['avg_cost'] += float(self.database[key]['_total_cost']) / self.database[key]['n_steps']


class Timer(object):
    def __init__(self):
        self.database = defaultdict(lambda: {'mode': 0, 'length': 0, 'total': 0})

    def start(self, key):
        self.database[key]['mode'] = 1
        self.database[key]['start'] = datetime.now()

    def end(self, key):
        assert self.database[key]['mode'] == 1, \
            'start timer for <{}> before ending'.format(key)

        self.database[key]['mode'] = 0
        self.database[key]['end'] = datetime.now()
        sub = self.database[key]['end'] - self.database[key]['start']
        self.database[key]['length'] = sub.microseconds
        self.database[key]['total'] += sub.microseconds

    def get(self, key):
        return self.database[key]['length']




''' Run a quick example pipeline and see if the
    pipeline can be loaded and saved correctly
    with Marmalade
'''

import os
import shutil
import unittest
import tempfile
import numpy as np

import sys
sys.path.append('..')
from timeseries_util import sarray


class TestSparseArray(unittest.TestCase):
    def test_sparsearray(self):
        fenceposts = np.array([0,4,10,22,27,28,30])
        num_dim = 3
        num_mix = 30
        array_2d = np.arange(num_dim*num_mix).reshape((num_dim, num_mix))

        sarray_2d = sarray(array_2d, fenceposts)

        self.assertEqual(sarray_2d._map(3, 2), 24)
        self.assertEqual(sarray_2d._map(0, 1), 1)
        self.assertEqual(sarray_2d._map(6, 2), 32)
        self.assertEqual(sum(sarray_2d.slice(([0], [2], [3]))), 13)
        self.assertEqual(
            list(sarray_2d.slice(([0, 1, 2], [0], [1,2,3])).flatten()),
            list(np.array([1,2,3,31,32,33,61,62,63])))
        self.assertEqual(sarray_2d.apply(np.sum), 4005)

    def test_sparsearray_dummy(self):
        array_3d = np.arange(12).reshape((2,3,2))
        array_2d = np.arange(12).reshape((2,6))
        fenceposts = np.array([0, 2, 4, 6])
        sarray_2d = sarray(array_2d, fenceposts)
        self.assertEqual(sum(sarray_2d.slice(([0], [0], [0]))), array_3d[0, 0, 0])
        self.assertEqual(list(sarray_2d.slice(([0], [0], [0,1]))), list(array_3d[0, 0, [0,1]]))
        self.assertEqual(sarray_2d.slice(([0,1], [0,1], [0,1,2])), array_3d)

        self.assertEqual(sarray_2d.slice((0, 0, 0)), 0)
        self.assertEqual(list(sarray_2d.slice((0, 0, [0]))), [0])
        self.assertEqual(list(sarray_2d.slice((0, 0, [[0]]))), [[0]])



import weights_parser
import scores
import cross_validation
import create_toy_data

import autograd_util
import common_util
import debug_util
import nn_util
import optimize

import compression
import distillation
import interpret
import models
import models_util
import emission_models
import pipeline

# infra
import datasets
import tests
import sample



from __future__ import absolute_import, print_function

import pdb
import os
import sys
from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.models import lr
from rlstm.weights_parser import WeightsParser
from rlstm.common_util import safe_log
from rlstm.nn_util import sigmoid

# HMM provides continuous support for HMM through
# continuous emission functions.
from rlstm.emission_models import bernoulli_multivariate
from rlstm.emission_models import diagonal_gaussian_multivariate

# ------------------------- #
#         HMM Utils         #
# ------------------------- #


def calc_loss_for_many_sequences(
        x_DM=None,
        y_CM=None,
        fenceposts_Np1=None,
        calc_log_proba_arr_for_x=None,
        eta_DK=None,
        start_log_proba_K=None,
        trans_log_proba_KK=None,
        output_weights_CK=None,
        return_state_array=False,
        **kwargs):
    ''' Compute loss (negative log probability) for many sequences given an HMM
    Args
    ----
    x_DM : 2D array, D x M, of real values
        Each row contains observed x data at one timestep of a sequence.
    y_CM : 2D array, C x M, of real values
        Each row contains observed y responses at one timestep of a sequence.
    fenceposts_Np1 : 1D array, N+1, of real values
        Fence posts for splitting time series into segments
    calc_log_proba_arr_for_x : function
        Computes log proba of each observation under each possible state
        Returns T x K array
    eta_DK : NumPy array, D x K (2K), of real values
        Contains parameters to calculate emission probabilities
    start_log_proba_K : 1D array, size K
        Starting-state log probabilities for the HMM.
        np.exp(start_log_proba_K) must sum to one
    trans_log_proba_KK : 2D array, size K x K
        Transition log probabilities for the HMM.
        np.exp(trans_log_proba_KK) must sum to one along each row
    output_weights_CK : 2D array, size C x K
        Logistic Regression weights to calculate output probabilities
    return_state_array: bool
        whether or not to return large array (loss will not need it)
    Returns
    -------
    loss_x : real scalar
    loss_y : real scalar
    '''
    n_seqs = fenceposts_Np1.size - 1
    n_dims = x_DM.shape[0]
    n_states = start_log_proba_K.shape[0]

    if return_state_array:
        belief_log_proba_NTK = np.zeros((x_DM.shape[1], n_states))
        n_i_track = 0
    else:
        belief_log_proba_NTK = None

    loss_x = 0.0
    loss_y = 0.0
    fenceposts_Np1_base = fenceposts_Np1[0]

    for n in xrange(n_seqs):
        # Get slice of x data corresp. to current sequence
        start_n = fenceposts_Np1[n] - fenceposts_Np1_base
        stop_n = fenceposts_Np1[n+1] - fenceposts_Np1_base
        x_n_DT = x_DM[:, start_n:stop_n]
        if y_CM is None:
            y_n_CT = None
        else:
            y_n_CT = y_CM[:, start_n:stop_n]

        loss_x_n, loss_y_n, belief_log_proba_TK = calc_loss_for_one_sequence(
            x_n_DT=x_n_DT,
            y_n_CT=y_n_CT,
            calc_log_proba_arr_for_x=calc_log_proba_arr_for_x,
            eta_DK=eta_DK,
            start_log_proba_K=start_log_proba_K,
            trans_log_proba_KK=trans_log_proba_KK,
            output_weights_CK=output_weights_CK,
            return_state_vector=return_state_array)

        loss_x = loss_x + loss_x_n
        loss_y = loss_y + loss_y_n

        if return_state_array:
            seq_n_time = belief_log_proba_TK.shape[0]
            belief_log_proba_NTK[n_i_track:n_i_track+seq_n_time] = belief_log_proba_TK
            n_i_track += seq_n_time

    return loss_x, loss_y, belief_log_proba_NTK


def calc_loss_for_one_sequence(
        x_n_DT=None,
        y_n_CT=None,
        calc_log_proba_arr_for_x=None,
        eta_DK=None,
        start_log_proba_K=None,
        trans_log_proba_KK=None,
        output_weights_CK=None,
        return_state_vector=False,
        **kwargs):
    ''' Compute loss (negative log probability) for one sequence given an HMM
    Args
    ----
    x_n_DT : 2D array, D x T, of real values
        Each row contains observed x data at timestep t of the sequence.
    y_n_CT : 2D array, C x T, of real values
        Each row contains observed y responses at timestep t of the seq.
    calc_log_proba_arr_for_x : function
        Computes log proba of each observation under each possible state
        Returns T x K array
    eta_DK : NumPy array, D x K, of real values
        Contains parameters to calculate emission probabilities
    start_log_proba_K : 1D array, size K
        Starting-state log probabilities for the HMM.
        np.exp(start_log_proba_K) must sum to one
    trans_log_proba_KK : 2D array, size K x K
        Transition log probabilities for the HMM.
        np.exp(trans_log_proba_KK) must sum to one along each row
    return_state_vector: bool
        whether or not to return large array (loss will not need it)
    Returns
    -------
    loss_x : real scalar
    loss_y : real scalar
    '''
    n_timesteps = x_n_DT.shape[1]
    n_states = start_log_proba_K.shape[0]

    if return_state_vector:
        belief_log_proba_TK = np.zeros((n_timesteps, n_states))
    else:
        belief_log_proba_TK = None

    # Compute log proba array
    x_n_log_proba_TK = calc_log_proba_arr_for_x(x_n_DT, eta_DK=eta_DK)
    # Initialize fwd belief vector at t = 0
    cur_belief_log_proba_K = start_log_proba_K + x_n_log_proba_TK[0]
    cur_x_log_proba = logsumexp(cur_belief_log_proba_K)
    cur_belief_log_proba_K = cur_belief_log_proba_K - cur_x_log_proba

    if return_state_vector:
        # store all state probabilities
        belief_log_proba_TK[0, :] = cur_belief_log_proba_K

    loss_x = cur_x_log_proba
    loss_y = 0.0

    if y_n_CT is not None:
        loss_y += compute_out_lprob(
            y_n_CT[:, 0],
            cur_belief_log_proba_K,
            output_weights_CK)

    # Recursively update forward beliefs via dynamic programming
    for t in range(1, n_timesteps):
        cur_belief_log_proba_K, cur_x_log_proba = update_belief_log_probas(
            cur_belief_log_proba_K,
            x_n_log_proba_TK[t],
            trans_log_proba_KK)

        if return_state_vector:
            belief_log_proba_TK[t, :] = cur_belief_log_proba_K

        # Compute log proba for x[t]
        loss_x += cur_x_log_proba

        # Compute log proba for y[t]
        if y_n_CT is not None:
            loss_y += compute_out_lprob(
                y_n_CT[:, t],
                cur_belief_log_proba_K,
                output_weights_CK)

    return loss_x, loss_y, belief_log_proba_TK


def update_belief_log_probas(
        prev_belief_log_proba_K,
        curr_data_log_proba_K,
        trans_log_proba_KK,
        return_norm_const=1):
    '''

    Examples
    --------
    # Uniform belief and uniform transitions, should just equal data probs
    >>> prev_belief_log_proba_K = np.log(np.asarray([0.5, 0.5]))
    >>> curr_data_log_proba_K = np.log(np.asarray([0.6, 0.4]))
    >>> trans_log_proba_KK = np.log(np.asarray([[0.5, 0.5], [0.5, 0.5]]))
    >>> curr_belief_log_proba_K = update_belief_log_probas(
    ...     prev_belief_log_proba_K, curr_data_log_proba_K, trans_log_proba_KK,
    ...     return_norm_const=0)
    >>> print(np.exp(curr_belief_log_proba_K))
    [ 0.6  0.4]

    >>> prev_belief_log_proba_K = np.log(np.asarray([0.5, 0.5]))
    >>> curr_data_log_proba_K = np.log(np.asarray([0.5, 0.5]))
    >>> trans_log_proba_KK = np.log(np.asarray([[0.9, 0.1], [0.9, 0.1]]))
    >>> curr_belief_log_proba_K = update_belief_log_probas(
    ...     prev_belief_log_proba_K, curr_data_log_proba_K, trans_log_proba_KK,
    ...     return_norm_const=0)
    >>> print(np.exp(curr_belief_log_proba_K))
    [ 0.9  0.1]
    '''
    cur_belief_log_proba_K = logsumexp(
        trans_log_proba_KK + prev_belief_log_proba_K[:, np.newaxis],
        axis=0)
    cur_belief_log_proba_K = cur_belief_log_proba_K \
        + curr_data_log_proba_K
    # Normalize in log space
    log_norm_const = logsumexp(cur_belief_log_proba_K)
    cur_belief_log_proba_K = cur_belief_log_proba_K - log_norm_const
    if return_norm_const:
        return cur_belief_log_proba_K, log_norm_const
    else:
        return cur_belief_log_proba_K


# Uses a sigmoid/logistic regression model for predicting the outputs
# given the input
# def compute_out_lprob(cur_output, lstate_norm, out_weights):
#     my_input = np.vstack((copy(lstate_norm), np.ones((1, lstate_norm.shape[1]))))
#     pred_set = sigmoid(np.dot(out_weights, my_input))
#     label_prob_set = pred_set * cur_output + \
#         (1 - pred_set) * (1 - cur_output)
#     lout = np.sum(safe_log(label_prob_set))
#     return lout

def compute_out_lprob(cur_output, lstate_norm, out_weights):
    my_input = np.hstack((copy(lstate_norm), [1]))
    pred_set = sigmoid(np.dot(out_weights, my_input))
    label_prob_set = pred_set * cur_output + \
        (1 - pred_set) * (1 - cur_output)
    lout = np.sum(safe_log(label_prob_set))
    return lout


# Computes the beliefs for a new set of data, given the weights and
# the observations (note that outputs aren't used in
# updating/computing the belief because we wish to respect the HMM
# structure of the model)

def get_state_prob_set(pred_fun, weights, obs_set, fence_set):
    lstate_set = pred_fun(weights, obs_set, fence_set)
    return lstate_set

# ------------------------- #
#     HMM Constructor       #
# ------------------------- #

def build(input_count, state_count, output_count=None, emission_type='bernoulli'):
    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)  lpi_mat is a
    # column vector, and so are all the lstates; ltrans( s, s') and
    # lobs( s, k ).
    # emission_type defines the distribution behind the emission
    # matrix used in HMM. discrete and continuous matrices are supported.
    assert emission_type in ['bernoulli', 'gaussian'], 'EMISSION_TYPE not recognized.'

    if emission_type == 'bernoulli':
        emission_module = bernoulli_multivariate
        emission_function = emission_module.calc_log_proba_data_arr_NK
    else:  # must be gaussian
        emission_module = diagonal_gaussian_multivariate
        emission_function = emission_module.calc_log_proba_data_arr_NK_flat

    parser = WeightsParser()
    parser.add_shape('lpi_mat', (state_count, 1))
    parser.add_shape('ltrans_mat', (state_count, state_count))

    if emission_type == 'bernoulli':
        parser.add_shape('eta_mat', (input_count, state_count))
    else:  # must be gaussian
        parser.add_shape('eta_mat', (input_count, state_count*2))

    if output_count is not None:
        parser.add_shape('out_weights', (output_count, state_count + 1))

    # Computes the state probabilities and overall log-likelihood for
    # the HMM.  lstate_set is the log of the belief at every time
    # step.  ll_vec is the log probs of the sequence of observation at
    # every time step: p(o1), p(o1,o2), p(o1,o2,o3), etc. mostly for
    # helping with diagnostics.  Somewhat arcane concatinations to get
    # around various autograd-related things.
    def outputs(weights, input_set, fence_set, output_set=None, return_lstate_set=False):
        # Get parameters
        lpi_mat = parser.get(weights, 'lpi_mat')
        ltrans_mat = parser.get(weights, 'ltrans_mat')
        eta_mat = parser.get(weights, 'eta_mat')  # see emission_models folder
        out_weights = None
        if output_set is not None:
            out_weights = parser.get(weights, 'out_weights')

        # Normalize the initial distribution and the transition distribution
        lpi_mat = lpi_mat - logsumexp(lpi_mat, axis=0)
        ltrans_mat = ltrans_mat - logsumexp(ltrans_mat, axis=1, keepdims=True)

        # initialize to 0
        in_ll, out_ll = 0, 0
        lstate_set = None

        # calculate loss for all sequences
        in_ll, out_ll, lstate_set_TK = calc_loss_for_many_sequences(
            x_DM=input_set,
            y_CM=output_set,
            fenceposts_Np1=fence_set,
            calc_log_proba_arr_for_x=emission_function,
            eta_DK=eta_mat,
            start_log_proba_K=lpi_mat.flatten(),
            trans_log_proba_KK=ltrans_mat,
            output_weights_CK=out_weights,
            return_state_array=return_lstate_set)

        if return_lstate_set:
            lstate_set = lstate_set_TK.T

        return in_ll, out_ll, lstate_set

    def log_prediction(weights, input_set, fence_set):
        ''' log probabilities '''
        out_weights = parser.get(weights, 'out_weights')
        _, _, lstate_set = outputs(weights, input_set, fence_set, return_lstate_set=True)

        logpreds = lr.logistic_log_predictions(
            out_weights,
            lstate_set,
            fence_set,
            state_count,
            output_count)

        return logpreds

    def generative_prediction(weights, input_set, fence_set):
        _, _, lstate_set = outputs(weights, input_set, fence_set, return_lstate_set=True)
        return lstate_set

    # Returns the log likelihood. If there are outputs, returns the
    # probability of the output set.  Otherwise, returns the
    # probability of the input set based on the model.
    def log_likelihood(weights, input_set, fence_set, output_set=None):
        in_ll, out_ll, _ = outputs(
            weights, input_set, fence_set, output_set) #, return_lstate_set=False)

        return in_ll if output_set is None else out_ll

    if not output_count is None:
        return log_prediction, log_likelihood, parser.num_weights
    else:
        return generative_prediction, log_likelihood, parser.num_weights
